import os
import math
import torch

import numpy as np
import torch.nn as nn

from abc import abstractmethod
from abc import ABC
from abc import ABCMeta
from copy import deepcopy
from tqdm import tqdm
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import TypeVar
from typing import Callable
from typing import Iterator
from typing import Optional
from typing import NamedTuple
from torch.optim import Optimizer
from cftool.misc import filter_kw
from cftool.misc import print_info
from cftool.misc import print_warning
from cftool.misc import lock_manager
from cftool.misc import check_requires
from cftool.misc import shallow_copy_dict
from cftool.misc import context_error_handler
from cftool.misc import get_num_positional_args
from cftool.misc import WithRegister
from cftool.array import to_numpy
from cftool.array import to_device
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type
from torch.optim.lr_scheduler import _LRScheduler

from .types import losses_type
from .types import configs_type
from .types import sample_weights_type
from .constants import LOSS_KEY
from .constants import INPUT_KEY
from .constants import LABEL_KEY
from .constants import PREDICTIONS_KEY
from .constants import BATCH_INDICES_KEY
from .constants import ORIGINAL_LABEL_KEY
from .misc.toolkit import eval_context
from .misc.toolkit import get_world_size
from .misc.toolkit import fix_denormal_states
from .misc.toolkit import ONNX

try:
    import onnx
except:
    onnx = None
try:
    from onnxsim import simplify as onnx_simplify

    def get_inputs(model: onnx.ModelProto) -> List[onnx.ValueInfoProto]:
        initializer_names = [x.name for x in model.graph.initializer]
        return [inp for inp in model.graph.input if inp.name not in initializer_names]

    def get_input_names(model: onnx.ModelProto) -> List[str]:
        input_names = [inp.name for inp in get_inputs(model)]
        return input_names

except:
    onnx_simplify = get_input_names = None  # type: ignore


dataset_dict: Dict[str, Type["IDataset"]] = {}
loader_dict: Dict[str, Type["IDataLoader"]] = {}
model_dict: Dict[str, Type["IDLModel"]] = {}
monitor_dict: Dict[str, Type["TrainerMonitor"]] = {}
loss_dict: Dict[str, Type["ILoss"]] = {}
multi_prefix_mapping: Dict[str, Type["MultiLoss"]] = {}
metric_dict: Dict[str, Type["_IMetric"]] = {}
callback_dict: Dict[str, Type["TrainerCallback"]] = {}

LossType = TypeVar("LossType", bound="ILoss", covariant=True)


# data


class IDataset(WithRegister["IDataset"], metaclass=ABCMeta):
    d = dataset_dict

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    @abstractmethod
    def __len__(self) -> int:
        pass


class IDataLoader(WithRegister["IDataLoader"], metaclass=ABCMeta):
    d = loader_dict
    data: IDataset
    batch_size: int

    def __init__(self, *, sample_weights: Optional[np.ndarray] = None):
        self.sample_weights = sample_weights

    @abstractmethod
    def __iter__(self) -> "IDataLoader":
        pass

    @abstractmethod
    def __next__(self) -> tensor_dict_type:
        pass

    @abstractmethod
    def disable_shuffle(self) -> None:
        pass

    @abstractmethod
    def recover_shuffle(self) -> None:
        pass

    def __len__(self) -> int:
        return math.ceil(len(self.data) / self.batch_size)

    def copy(self) -> "IDataLoader":
        return deepcopy(self)

    def temporarily_disable_shuffle(self) -> context_error_handler:
        class _(context_error_handler):
            def __init__(self, loader: IDataLoader):
                self.loader = loader

            def __enter__(self) -> None:
                self.loader.disable_shuffle()

            def _normal_exit(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                self.loader.recover_shuffle()

        return _(self)


class IDataModule(ABC):
    id_file: str
    info_name: str
    data_folder: str
    package_folder: str

    @property
    @abstractmethod
    def info(self) -> Dict[str, Any]:
        pass

    @abstractmethod
    def prepare(self, sample_weights: sample_weights_type) -> None:
        pass

    @abstractmethod
    def initialize(self) -> Any:
        pass

    @abstractmethod
    def save_info(self, folder: str) -> None:
        pass

    @abstractmethod
    def save(self, folder: str) -> None:
        pass


# model


def _forward(
    m: nn.Module,
    batch_idx: int,
    batch: tensor_dict_type,
    general_input_key: str,
    state: Optional["TrainerState"] = None,
    *,
    general_output_key: str = PREDICTIONS_KEY,
    **kwargs: Any,
) -> tensor_dict_type:
    fn = m.forward
    if check_requires(fn, "general_output_key"):
        kwargs["general_output_key"] = general_output_key
    kw = filter_kw(fn, kwargs)
    args: List[Any] = []
    if check_requires(fn, "batch_idx"):
        args.append(batch_idx)
    if get_num_positional_args(fn) > 0:
        args.append(batch if check_requires(fn, "batch") else batch[general_input_key])
    if check_requires(fn, "state"):
        args.append(state)
    rs = m(*args, **kw)
    if not isinstance(rs, dict):
        rs = {general_output_key: rs}
    return rs


class WithDeviceMixin:
    parameters: Callable[["WithDeviceMixin"], Iterator[nn.Parameter]]

    @property
    def device(self) -> torch.device:
        params = list(self.parameters())  # type: ignore
        return torch.device("cpu") if not params else params[0].device


class IDLModel(
    WithDeviceMixin,
    nn.Module,
    WithRegister["IDLModel"],
    metaclass=ABCMeta,
):
    d = model_dict

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__()

    def _init_with_trainer(self, trainer: "ITrainer") -> None:
        pass

    @abstractmethod
    def forward(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        state: Optional["TrainerState"] = None,
        **kwargs: Any,
    ) -> tensor_dict_type:
        pass

    def onnx_forward(self, batch: tensor_dict_type) -> Any:
        return self.forward(0, batch)

    def summary_forward(self, batch_idx: int, batch: tensor_dict_type) -> None:
        self.forward(batch_idx, batch)

    def to_onnx(
        self,
        export_folder: str,
        input_sample: tensor_dict_type,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        onnx_file: str = "model.onnx",
        opset: int = 11,
        simplify: bool = True,
        forward_fn: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        output_names: Optional[List[str]] = None,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> "IDLModel":
        # prepare
        device = self.device
        model = self.cpu()
        if num_samples is not None:
            input_sample = {k: v[:num_samples] for k, v in input_sample.items()}
        onnx_forward = forward_fn or model.onnx_forward
        input_names = sorted(input_sample.keys())
        if output_names is None:
            if forward_fn is not None:
                msg = "`output_names` should be provided when `forward_fn` is provided"
                raise ValueError(msg)
            with eval_context(model):
                forward_results = onnx_forward(shallow_copy_dict(input_sample))
            if not isinstance(forward_results, dict):
                forward_results = {PREDICTIONS_KEY: forward_results}
            output_names = sorted(forward_results.keys())
        # setup
        kwargs = shallow_copy_dict(kwargs)
        kwargs["input_names"] = input_names
        kwargs["output_names"] = output_names
        kwargs["opset_version"] = opset
        kwargs["export_params"] = True
        kwargs["do_constant_folding"] = True
        if dynamic_axes is None:
            dynamic_axes = {}
        elif isinstance(dynamic_axes, list):
            dynamic_axes = {axis: f"axis.{axis}" for axis in dynamic_axes}
        if num_samples is None:
            dynamic_axes[0] = "batch_size"
        dynamic_axes_settings = {}
        for name in input_names + output_names:
            dynamic_axes_settings[name] = dynamic_axes
        kwargs["dynamic_axes"] = dynamic_axes_settings
        kwargs["verbose"] = verbose
        # export
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)

        class ONNXWrapper(torch.nn.Module):
            def __init__(self) -> None:
                super().__init__()
                self.model = model

            def forward(self, batch: Dict[str, Any]) -> Dict[str, Any]:
                rs = onnx_forward(batch)
                if isinstance(rs, torch.Tensor):
                    return {k: rs for k in output_names}  # type: ignore
                return {k: rs[k] for k in output_names}  # type: ignore

        with lock_manager(base_folder, []) as lock:
            onnx_path = os.path.join(export_folder, onnx_file)
            lock._stuffs = [onnx_path]
            os.makedirs(export_folder, exist_ok=True)
            m_onnx = ONNXWrapper()
            original_states = model.state_dict()
            fixed_states = fix_denormal_states(original_states, verbose=verbose)
            with eval_context(m_onnx):
                model.load_state_dict(fixed_states)
                torch.onnx.export(
                    m_onnx,
                    ({k: input_sample[k] for k in input_names}, {}),
                    onnx_path,
                    **shallow_copy_dict(kwargs),
                )
                model.load_state_dict(original_states)
                if not simplify:
                    return self.to(device)
                if onnx is None:
                    print_warning(
                        "`onnx` is not installed, "
                        "so the exported onnx model will not be simplified"
                    )
                    return self.to(device)
                if onnx_simplify is None or get_input_names is None:
                    print_warning(
                        "`onnx-simplifier` is not installed, "
                        "so the exported onnx model will not be simplified"
                    )
                    return self.to(device)
                try:
                    onnx_model = onnx.load(onnx_path)
                    final_input_names = get_input_names(onnx_model)
                    model_simplified, check = onnx_simplify(
                        onnx_model,
                        test_input_shapes={
                            name: tensor.shape
                            for name, tensor in input_sample.items()
                            if name in final_input_names
                        },
                    )
                except Exception as err:
                    if verbose:
                        print_warning(f"Failed to simplify ONNX model ({err})")
                    model_simplified = None
                    check = False
                if verbose:
                    tag = " " if check else " not "
                    print_info(f"Simplified ONNX model is{tag}validated!")
                if check and model_simplified is not None:
                    onnx.save(model_simplified, onnx_path)
        return self.to(device)


class StepOutputs(NamedTuple):
    forward_results: tensor_dict_type
    loss_dict: Dict[str, float]


class MetricsOutputs(NamedTuple):
    final_score: float
    metric_values: Dict[str, float]


class InferenceOutputs(NamedTuple):
    forward_results: np_dict_type
    labels: Optional[np.ndarray]
    metric_outputs: Optional[MetricsOutputs]
    loss_items: Optional[Dict[str, float]]


class ModelWithCustomSteps(IDLModel, metaclass=ABCMeta):
    custom_train_step: bool = True
    custom_evaluate_step: bool = True
    custom_params_groups: bool = False
    custom_ddp_initialization: bool = False

    @abstractmethod
    def train_step(
        self,
        batch_idx: int,
        batch: tensor_dict_type,
        trainer: "ITrainer",
        forward_kwargs: Dict[str, Any],
        loss_kwargs: Dict[str, Any],
    ) -> StepOutputs:
        pass

    # TODO : Add `forward_kwargs`
    @abstractmethod
    def evaluate_step(
        self,
        loader: IDataLoader,
        portion: float,
        trainer: "ITrainer",
    ) -> MetricsOutputs:
        pass

    def params_groups(self, m: nn.Module) -> Any:
        pass

    def init_ddp(self) -> None:
        pass

    def permute_trainer_config(self, trainer_config: Dict[str, Any]) -> None:
        pass


# trainer types


class TrainerState:
    def __init__(
        self,
        loader: IDataLoader,
        *,
        num_epoch: int,
        max_epoch: int,
        fixed_steps: Optional[int] = None,
        extension: int = 5,
        enable_logging: bool = True,
        min_num_sample: int = 3000,
        snapshot_start_step: Optional[int] = None,
        max_snapshot_file: int = 5,
        num_snapshot_per_epoch: int = 2,
        num_step_per_log: int = 350,
        num_step_per_snapshot: Optional[int] = None,
        max_step_per_snapshot: int = 1000,
        min_snapshot_epoch_gap: int = 0,
    ):
        self.step = self.epoch = 0
        self.batch_size = loader.batch_size * get_world_size()
        self.num_step_per_epoch = len(loader)
        self.num_epoch = num_epoch
        self.max_epoch = max_epoch
        self.fixed_steps = fixed_steps
        self.extension = extension
        self.enable_logging = enable_logging
        self.min_num_sample = min_num_sample
        if snapshot_start_step is None:
            snapshot_start_step = math.ceil(min_num_sample / self.batch_size)
        self.snapshot_start_step = snapshot_start_step
        self.max_snapshot_file = max_snapshot_file
        self.num_snapshot_per_epoch = num_snapshot_per_epoch
        self.num_step_per_log = num_step_per_log
        if num_step_per_snapshot is None:
            num_step_per_snapshot = max(1, int(len(loader) / num_snapshot_per_epoch))
            num_step_per_snapshot = min(max_step_per_snapshot, num_step_per_snapshot)
        self.num_step_per_snapshot = num_step_per_snapshot
        self.max_step_per_snapshot = max_step_per_snapshot
        self.min_snapshot_epoch_gap = min_snapshot_epoch_gap
        self._previous_snapshot_epoch = 0

    def set_terminate(self) -> None:
        self.step = self.epoch = -1

    def update_snapshot_epoch(self) -> None:
        self._previous_snapshot_epoch = self.epoch

    @property
    def config(self) -> Dict[str, Any]:
        return {
            "num_epoch": self.num_epoch,
            "max_epoch": self.max_epoch,
            "fixed_steps": self.fixed_steps,
            "extension": self.extension,
            "enable_logging": self.enable_logging,
            "min_num_sample": self.min_num_sample,
            "snapshot_start_step": self.snapshot_start_step,
            "max_snapshot_file": self.max_snapshot_file,
            "num_snapshot_per_epoch": self.num_snapshot_per_epoch,
            "num_step_per_log": self.num_step_per_log,
            "num_step_per_snapshot": self.num_step_per_snapshot,
            "max_step_per_snapshot": self.max_step_per_snapshot,
        }

    @property
    def is_terminate(self) -> bool:
        return self.epoch == -1

    @property
    def should_train(self) -> bool:
        if self.fixed_steps is not None:
            return self.step < self.fixed_steps
        return self.epoch < self.num_epoch

    @property
    def should_terminate(self) -> bool:
        if self.fixed_steps is None:
            return False
        return self.step == self.fixed_steps

    @property
    def should_monitor(self) -> bool:
        return self.step % self.num_step_per_snapshot == 0

    @property
    def should_log_lr(self) -> bool:
        if not self.enable_logging:
            return False
        denominator = min(self.num_step_per_epoch, 10)
        return self.step % denominator == 0

    @property
    def should_log_losses(self) -> bool:
        if not self.enable_logging:
            return False
        patience = max(4, int(round(self.num_step_per_epoch / 50.0)))
        denominator = min(self.num_step_per_epoch, patience)
        return self.step % denominator == 0

    @property
    def should_log_artifacts(self) -> bool:
        return self.should_log_metrics_msg

    @property
    def should_log_metrics_msg(self) -> bool:
        if not self.enable_logging:
            return False
        if self.is_terminate:
            return True
        min_period = math.ceil(self.num_step_per_log / self.num_step_per_snapshot)
        period = max(1, int(min_period)) * self.num_step_per_snapshot
        return self.step % period == 0

    @property
    def can_snapshot(self) -> bool:
        if self.is_terminate:
            return True
        return self.epoch - self._previous_snapshot_epoch >= self.min_snapshot_epoch_gap

    @property
    def should_start_snapshot(self) -> bool:
        return self.step >= self.snapshot_start_step

    @property
    def should_extend_epoch(self) -> bool:
        return self.epoch == self.num_epoch and self.epoch < self.max_epoch

    @property
    def reached_max_epoch(self) -> bool:
        return self.epoch > self.max_epoch

    @property
    def disable_logging(self) -> context_error_handler:
        class _(context_error_handler):
            def __init__(self, state: TrainerState):
                self.state = state
                self.enabled = state.enable_logging

            def __enter__(self) -> None:
                self.state.enable_logging = False

            def _normal_exit(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
                self.state.enable_logging = self.enabled

        return _(self)


class TrainerMonitor(WithRegister["TrainerMonitor"], metaclass=ABCMeta):
    d = monitor_dict

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    @abstractmethod
    def snapshot(self, new_score: float) -> bool:
        pass

    @abstractmethod
    def check_terminate(self, new_score: float) -> bool:
        pass

    @abstractmethod
    def punish_extension(self) -> None:
        pass

    def handle_extension(self, state: TrainerState) -> None:
        if state.should_extend_epoch:
            self.punish_extension()
            new_epoch = state.num_epoch + state.extension
            state.num_epoch = min(new_epoch, state.max_epoch)


class MonitorResults(NamedTuple):
    terminate: bool
    save_checkpoint: bool
    metric_outputs: Optional[MetricsOutputs]


# loss


class ILoss(nn.Module, WithRegister[LossType], metaclass=ABCMeta):
    d = loss_dict
    placeholder_key = "[PLACEHOLDER]"

    def __init__(self, reduction: str = "mean", **kwargs: Any):
        super().__init__()
        self.config = kwargs
        self._init_config()
        self._reduction = reduction

    def _init_config(self) -> None:
        pass

    def _reduce(self, losses: torch.Tensor) -> torch.Tensor:
        if self._reduction == "none":
            return losses
        if self._reduction == "mean":
            return losses.mean()
        if self._reduction == "sum":
            return losses.sum()
        raise NotImplementedError(f"reduction '{self._reduction}' is not implemented")

    @abstractmethod
    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        # return losses without reduction
        pass

    def forward(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
    ) -> tensor_dict_type:
        losses = self._core(forward_results, batch, state)
        if isinstance(losses, torch.Tensor):
            return {LOSS_KEY: self._reduce(losses)}
        # requires returns a value with LOSS_KEY as its key
        return {k: self._reduce(v) for k, v in losses.items()}

    @classmethod
    def parse(cls, name: str) -> str:
        if ":" not in name:
            return name
        split = name.split(":")
        if split[-2] != AuxLoss.identifier:
            for prefix, base in multi_prefix_mapping.items():
                if name.startswith(f"{prefix}:"):
                    loss_names = name.split(":")[1].split(",")
                    return base.register_(loss_names)
            return name
        loss_name = cls.parse(":".join(split[:-2]))
        aux_names = split[-1].split(",")
        return AuxLoss.register_(loss_name, aux_names)


class MultiLoss(ILoss, metaclass=ABCMeta):
    prefix: str

    names: Union[str, List[str]]
    base_losses: nn.ModuleList

    def _init_config(self) -> None:
        base_losses: List[ILoss]
        if isinstance(self.names, str):
            base_losses = [ILoss.make(self.names, self.config)]
        else:
            base_losses = [
                ILoss.make(name, self.config.get(name, {})) for name in self.names
            ]
        self.base_losses = nn.ModuleList(base_losses)

    @staticmethod
    def _inject(key: str, base_losses: losses_type, all_losses: losses_type) -> None:
        if isinstance(base_losses, dict):
            base_losses = base_losses[LOSS_KEY]
        all_losses[key] = base_losses

    @classmethod
    def register_(
        cls,
        base_loss_names: Union[str, List[str]],
        *,
        tag: Optional[str] = None,
    ) -> str:
        if tag is None:
            if isinstance(base_loss_names, str):
                tag = f"{cls.prefix}_{base_loss_names}"
            else:
                tag = f"{cls.prefix}_{'_'.join(base_loss_names)}"
        if tag in cls.d:
            return tag

        @cls.register(tag)  # type: ignore
        class _(cls):  # type: ignore
            names = base_loss_names

        return tag

    @classmethod
    def record_prefix(cls) -> Callable[[Type["MultiLoss"]], Type["MultiLoss"]]:
        def _(cls_: Type[MultiLoss]) -> Type[MultiLoss]:
            global multi_prefix_mapping
            multi_prefix_mapping[cls_.prefix] = cls_
            return cls_

        return _


class AuxLoss(ILoss):
    identifier: str = "aux"
    main_loss_key: str = "main"

    loss: ILoss
    loss_name: str
    aux_names: List[str]

    def _init_config(self) -> None:
        self.loss = ILoss.make(self.loss_name, self.config)

    @staticmethod
    def _convert(losses: losses_type) -> torch.Tensor:
        if isinstance(losses, dict):
            return losses[LOSS_KEY]
        return losses

    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        main_losses = self.loss._core(forward_results, batch, state, **kwargs)
        losses = {self.main_loss_key: self._convert(main_losses)}
        for name in self.aux_names:
            losses[name] = self._convert(
                self.loss._core(
                    {PREDICTIONS_KEY: forward_results[name]},
                    {LABEL_KEY: batch[name]},
                    state,
                    **kwargs,
                )
            )
        losses[LOSS_KEY] = sum(losses.values())
        return losses

    @classmethod
    def register_(
        cls,
        base_loss_name: str,
        aux_loss_names: List[str],
        *,
        tag: Optional[str] = None,
    ) -> str:
        for name in aux_loss_names:
            if name == cls.main_loss_key:
                raise ValueError(f"should not use '{cls.main_loss_key}' as aux name")
        if tag is None:
            tag = f"{base_loss_name}_{cls.identifier}_{'_'.join(aux_loss_names)}"
        if tag in cls.d:
            return tag

        @cls.register(tag)  # type: ignore
        class _(cls):  # type: ignore
            loss_name = base_loss_name
            aux_names = aux_loss_names

        return tag


@ILoss.register(ILoss.placeholder_key)
class PlaceholderLoss(ILoss):
    def _core(
        self,
        forward_results: tensor_dict_type,
        batch: tensor_dict_type,
        state: Optional[TrainerState] = None,
        **kwargs: Any,
    ) -> losses_type:
        raise ValueError("`forward` should not be called in `PlaceholderLoss`")


# inference


class IInference:
    def __init__(
        self,
        *,
        onnx: Optional[ONNX] = None,
        model: Optional[IDLModel] = None,
        use_grad_in_predict: bool = False,
    ):
        self.onnx = onnx
        self.model = model
        if onnx is None and model is None:
            raise ValueError("either `onnx` or `model` should be provided")
        if onnx is not None and model is not None:
            raise ValueError("only one of `onnx` & `model` should be provided")
        self.use_grad_in_predict = use_grad_in_predict

    def get_outputs(
        self,
        loader: IDataLoader,
        *,
        portion: float = 1.0,
        state: Optional[TrainerState] = None,
        metrics: Optional["_IMetric"] = None,
        loss: Optional[ILoss] = None,
        return_outputs: bool = True,
        stack_outputs: bool = True,
        use_tqdm: bool = False,
        **kwargs: Any,
    ) -> InferenceOutputs:
        def _core() -> InferenceOutputs:
            results: Dict[str, Optional[List[np.ndarray]]] = {}
            metric_outputs_list: List[MetricsOutputs] = []
            loss_items: Dict[str, List[float]] = {}
            labels: Dict[str, List[np.ndarray]] = {}
            iterator = enumerate(loader)
            if use_tqdm:
                iterator = tqdm(iterator, "inference", len(loader))
            requires_all_outputs = return_outputs
            if metrics is not None and metrics.requires_all:
                requires_all_outputs = True
            for i, batch in iterator:
                if i / len(loader) >= portion:
                    break
                np_batch = {
                    batch_key: None if batch_tensor is None else to_numpy(batch_tensor)
                    for batch_key, batch_tensor in batch.items()
                }
                if self.model is not None:
                    batch = to_device(batch, self.model.device)
                if self.onnx is not None:
                    local_outputs = self.onnx.predict(np_batch)
                else:
                    assert self.model is not None
                    with eval_context(self.model, use_grad=use_grad):
                        assert not self.model.training
                        local_outputs = self.model(
                            i,
                            batch,
                            state,
                            **shallow_copy_dict(kwargs),
                        )
                # gather outputs
                requires_metrics = metrics is not None and not metrics.requires_all
                requires_np = requires_metrics or requires_all_outputs
                np_outputs: np_dict_type = {}
                for k, v in local_outputs.items():
                    if not requires_np:
                        results[k] = None
                        continue
                    if v is None:
                        continue
                    if isinstance(v, np.ndarray):
                        v_np = v
                    elif isinstance(v, torch.Tensor):
                        v_np = to_numpy(v)
                    elif isinstance(v, list):
                        if isinstance(v[0], np.ndarray):
                            v_np = v
                        else:
                            v_np = list(map(to_numpy, v))
                    else:
                        raise ValueError(f"unrecognized value ({k}={type(v)}) occurred")
                    np_outputs[k] = v_np
                    if not requires_all_outputs:
                        results[k] = None
                    else:
                        results.setdefault(k, []).append(v_np)  # type: ignore
                if requires_np:
                    for k, v in batch.items():
                        if (
                            k == INPUT_KEY
                            or k == ORIGINAL_LABEL_KEY
                            or k.endswith(BATCH_INDICES_KEY)
                        ):
                            continue
                        if v is None:
                            continue
                        if k != LABEL_KEY and len(v.shape) > 2:
                            continue
                        if not isinstance(v, np.ndarray):
                            v = to_numpy(v)
                        labels.setdefault(k, []).append(v)
                # metrics
                if requires_metrics:
                    sub_outputs = metrics.evaluate(np_batch, np_outputs, loader)  # type: ignore
                    metric_outputs_list.append(sub_outputs)
                # loss
                if loss is not None:
                    with eval_context(loss, use_grad=use_grad):
                        local_losses = loss(local_outputs, batch)
                    for k, v in local_losses.items():
                        loss_items.setdefault(k, []).append(v.item())
            # gather outputs
            final_results: Dict[str, Union[np.ndarray, Any]]
            if not requires_all_outputs:
                final_results = {k: None for k in results}
            else:
                final_results = {
                    batch_key: batch_results
                    if not stack_outputs
                    else np.vstack(batch_results)
                    if isinstance(batch_results[0], np.ndarray)
                    else [
                        np.vstack([batch[i] for batch in batch_results])
                        for i in range(len(batch_results[0]))
                    ]
                    for batch_key, batch_results in results.items()
                    if batch_results is not None
                }
            # gather metric outputs
            if metrics is None:
                metric_outputs = None
            elif metrics.requires_all:
                metric_outputs = metrics.evaluate(
                    {k: np.vstack(v) for k, v in labels.items()},
                    final_results,
                    loader,
                )
            else:
                scores = []
                metric_values: Dict[str, List[float]] = {}
                for sub_outputs in metric_outputs_list:
                    scores.append(sub_outputs.final_score)
                    for k, v in sub_outputs.metric_values.items():
                        metric_values.setdefault(k, []).append(v)
                metric_outputs = MetricsOutputs(
                    sum(scores) / len(scores),
                    {k: sum(vl) / len(vl) for k, vl in metric_values.items()},
                )

            target_labels = labels.get(LABEL_KEY, [])
            return InferenceOutputs(
                final_results,
                None if not target_labels else np.vstack(target_labels),
                metric_outputs,
                None
                if not loss_items
                else {k: sum(v) / len(v) for k, v in loss_items.items()},
            )

        use_grad = kwargs.pop("use_grad", self.use_grad_in_predict)
        with loader.temporarily_disable_shuffle():
            try:
                return _core()
            except:
                use_grad = self.use_grad_in_predict = True
                return _core()


# metrics


class _IMetric(WithRegister["_IMetric"], metaclass=ABCMeta):
    d = metric_dict

    trainer: "ITrainer"

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    @property
    @abstractmethod
    def is_positive(self) -> bool:
        pass

    @abstractmethod
    def _core(
        self,
        np_batch: np_dict_type,
        np_outputs: np_dict_type,
        loader: Optional[IDataLoader],
    ) -> float:
        pass

    @property
    def requires_all(self) -> bool:
        return False

    @classmethod
    def fuse(
        cls,
        names: Union[str, List[str]],
        configs: configs_type = None,
        *,
        metric_weights: Optional[Dict[str, float]] = None,
    ) -> "_IMetric":
        metrics = _IMetric.make_multiple(names, configs)
        if isinstance(metrics, _IMetric):
            return metrics
        return _MultipleMetrics(metrics, weights=metric_weights)

    def evaluate(
        self,
        np_batch: np_dict_type,
        np_outputs: np_dict_type,
        loader: Optional[IDataLoader] = None,
    ) -> MetricsOutputs:
        metric = self._core(np_batch, np_outputs, loader)
        score = metric * (1.0 if self.is_positive else -1.0)
        return MetricsOutputs(score, {self.__identifier__: metric})


class _MultipleMetrics(_IMetric):
    @property
    def is_positive(self) -> bool:
        raise NotImplementedError

    @property
    def requires_all(self) -> bool:
        return any(metric.requires_all for metric in self.metrics)

    def _core(
        self,
        np_batch: np_dict_type,
        np_outputs: np_dict_type,
        loader: Optional[IDataLoader],
    ) -> float:
        raise NotImplementedError

    def __init__(
        self,
        metric_list: List[_IMetric],
        *,
        weights: Optional[Dict[str, float]] = None,
    ):
        super().__init__()
        self.metrics = metric_list
        self.weights = weights or {}
        self.__identifier__ = " | ".join(m.__identifier__ for m in metric_list)

    def evaluate(
        self,
        np_batch: np_dict_type,
        np_outputs: np_dict_type,
        loader: Optional[IDataLoader] = None,
    ) -> MetricsOutputs:
        scores: List[float] = []
        weights: List[float] = []
        metrics_values: Dict[str, float] = {}
        for metric in self.metrics:
            metric_outputs = metric.evaluate(np_batch, np_outputs, loader)
            w = self.weights.get(metric.__identifier__, 1.0)
            weights.append(w)
            scores.append(metric_outputs.final_score * w)
            metrics_values.update(metric_outputs.metric_values)
        return MetricsOutputs(sum(scores) / sum(weights), metrics_values)


# trainer interface


class DeviceInfo(NamedTuple):
    cuda: Optional[str]
    rank: Optional[int]

    @property
    def device(self) -> torch.device:
        if self.rank is not None:
            return torch.device(f"cuda:{self.rank}")
        return torch.device("cpu" if self.cuda is None else f"cuda:{self.cuda}")


class OptimizerPack(NamedTuple):
    scope: str
    optimizer_name: str
    scheduler_name: Optional[str] = None
    optimizer_config: Optional[Dict[str, Any]] = None
    scheduler_config: Optional[Dict[str, Any]] = None


class TqdmSettings(NamedTuple):
    use_tqdm: bool = False
    use_step_tqdm: bool = False
    use_tqdm_in_validation: bool = False
    in_distributed: bool = False
    position: int = 0
    desc: str = "epoch"


class TrainerCallback(WithRegister["TrainerCallback"]):
    d = callback_dict
    is_rank_0: bool = True

    def __init__(self, *args: Any, **kwargs: Any):
        pass

    def initialize(self) -> None:
        pass

    def mutate_train_forward_kwargs(
        self,
        kwargs: Dict[str, Any],
        trainer: "ITrainer",
    ) -> None:
        pass

    def mutate_train_loss_kwargs(
        self,
        kwargs: Dict[str, Any],
        trainer: "ITrainer",
    ) -> None:
        pass

    def before_loop(self, trainer: "ITrainer") -> None:
        pass

    def log_lr(self, key: str, lr: float, state: TrainerState) -> None:
        pass

    def log_metrics(self, metric_outputs: MetricsOutputs, state: TrainerState) -> None:
        pass

    def log_metrics_msg(
        self,
        metric_outputs: MetricsOutputs,
        metrics_log_path: str,
        state: TrainerState,
    ) -> None:
        pass

    def log_artifacts(self, trainer: "ITrainer") -> None:
        pass

    def after_step(self, step_outputs: StepOutputs, state: TrainerState) -> None:
        pass

    def after_monitor(
        self,
        monitor_results: MonitorResults,
        state: TrainerState,
    ) -> None:
        pass

    def finalize(self, trainer: "ITrainer") -> None:
        pass


class ITrainer(ABC):
    loss: ILoss
    model: IDLModel
    metrics: Optional[_IMetric]
    monitors: List[TrainerMonitor]
    callbacks: List[TrainerCallback]
    optimizers: Dict[str, Optimizer]
    schedulers: Dict[str, Optional[_LRScheduler]]

    state: TrainerState
    device_info: DeviceInfo
    train_loader: IDataLoader
    train_loader_copy: IDataLoader
    valid_loader: Optional[IDataLoader]
    inference: IInference

    workplace: str
    checkpoint_folder: Optional[str]

    tqdm_settings: TqdmSettings
    state_config: Dict[str, Any]
    num_epoch: int
    max_epoch: int
    fixed_steps: Optional[int]
    valid_portion: float
    use_amp: bool
    use_zero: bool
    clip_norm: float
    grad_scaler: torch.cuda.amp.GradScaler

    @property
    @abstractmethod
    def config(self) -> Dict[str, Any]:
        pass

    @property
    @abstractmethod
    def device(self) -> torch.device:
        pass

    @property
    @abstractmethod
    def use_tqdm_in_validation(self) -> bool:
        pass

    @property
    @abstractmethod
    def validation_loader(self) -> IDataLoader:
        pass

    @property
    @abstractmethod
    def input_sample(self) -> tensor_dict_type:
        pass

    @property
    @abstractmethod
    def has_checkpoint_folder(self) -> bool:
        pass

    @property
    @abstractmethod
    def model_has_custom_steps(self) -> bool:
        pass

    @property
    @abstractmethod
    def model_for_training(self) -> nn.Module:
        pass

    # init

    @abstractmethod
    def _init_ddp(self) -> None:
        pass

    @abstractmethod
    def _init_finetune(self) -> None:
        pass

    @abstractmethod
    def default_lr_configs(
        self,
        optimizer: Optimizer,
        optimizer_config: Dict[str, Any],
    ) -> Dict[str, Dict[str, Any]]:
        pass

    @abstractmethod
    def _define_optimizer(self, pack: OptimizerPack) -> Optimizer:
        pass

    @abstractmethod
    def _define_scheduler(self, optimizer: Optimizer, pack: OptimizerPack) -> None:
        pass

    @abstractmethod
    def _init_optimizers(self) -> None:
        pass

    # core

    @abstractmethod
    def post_loss_step(self, loss_dict: tensor_dict_type) -> None:
        pass

    @abstractmethod
    def weighted_loss_score(self, loss_items: Dict[str, float]) -> float:
        pass

    @abstractmethod
    def clip_norm_step(self) -> None:
        pass

    @abstractmethod
    def optimizer_step(self) -> None:
        pass

    @abstractmethod
    def scheduler_step(self) -> None:
        pass

    @abstractmethod
    def _get_scheduler_settings(
        self,
        key: str,
        scheduler: Any,
    ) -> Tuple[bool, Dict[str, Any]]:
        pass

    @abstractmethod
    def _logging_step(self, metrics_outputs: MetricsOutputs) -> None:
        pass

    @abstractmethod
    def _monitor_step(self, step_outputs: StepOutputs) -> MonitorResults:
        pass

    @abstractmethod
    def _step(self, batch_idx: int, batch: tensor_dict_type) -> StepOutputs:
        pass

    @abstractmethod
    def fit(
        self,
        data: IDataModule,
        loss: ILoss,
        model: IDLModel,
        inference: IInference,
        *,
        config_export_file: Optional[str] = None,
        show_summary: Optional[bool] = None,
        cuda: Optional[str] = None,
    ) -> "ITrainer":
        pass

    @abstractmethod
    def get_metrics(
        self,
        *,
        portion: float = 1.0,
        loader: Optional[IDataLoader] = None,
    ) -> MetricsOutputs:
        pass

    @abstractmethod
    def save_checkpoint(
        self,
        score: float,
        folder: Optional[str] = None,
        *,
        no_history: bool = False,
    ) -> None:
        pass

    @abstractmethod
    def restore_checkpoint(
        self,
        folder: str = None,
        strict: bool = True,
        state_dict_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
    ) -> bool:
        pass


__all__ = [
    "_forward",
    "dataset_dict",
    "loader_dict",
    "loss_dict",
    "multi_prefix_mapping",
    "metric_dict",
    "callback_dict",
    "IDataset",
    "IDataLoader",
    "IDLModel",
    "ModelWithCustomSteps",
    "StepOutputs",
    "TrainerState",
    "TrainerMonitor",
    "MonitorResults",
    "ILoss",
    "MultiLoss",
    "AuxLoss",
    "ONNX",
    "InferenceOutputs",
    "IInference",
    "MetricsOutputs",
    "_IMetric",
    "_MultipleMetrics",
    "DeviceInfo",
    "OptimizerPack",
    "TqdmSettings",
    "TrainerCallback",
    "ITrainer",
]



import torch

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Optional

from .protocol import ISampler
from .protocol import IDiffusion
from .protocol import DDPMQSampler
from ..utils import extract_to
from ..utils import get_timesteps


@ISampler.register("basic")
class BasicSampler(ISampler):
    def __init__(
        self,
        model: IDiffusion,
        *,
        temperature: float = 1.0,
        default_steps: Optional[int] = None,
    ):
        super().__init__(model)
        if default_steps is None:
            default_steps = model.t
        self.temperature = temperature
        self.default_steps = default_steps

    @property
    def q_sampler(self) -> DDPMQSampler:
        return self.model.q_sampler

    @property
    def sample_kwargs(self) -> Dict[str, Any]:
        return dict(temperature=self.temperature)

    def sample_step(
        self,
        image: Tensor,
        cond: Optional[Tensor],
        step: int,
        total_step: int,
        *,
        temperature: float = 1.0,
        **kwargs: Any,
    ) -> Tensor:
        shape = image.shape
        num_dim = len(shape)
        ts = get_timesteps(total_step - step - 1, shape[0], image.device)
        net = self.model.denoise(image, ts, cond)
        parameterization = self.model.parameterization
        if parameterization == "eps":
            coef1 = extract_to(self.model.posterior_coef1, ts, num_dim)
            coef2 = extract_to(self.model.posterior_coef2, ts, num_dim)
            mean = coef1 * (image - coef2 * net)
        elif parameterization == "x0":
            mean = net
        else:
            msg = f"unrecognized parameterization '{parameterization}' occurred"
            raise NotImplementedError(msg)
        noise = torch.randn_like(image) * temperature
        noise_mask_shape = shape[0], *((1,) * (num_dim - 1))
        noise_mask = (1.0 - (ts == 0).float()).view(noise_mask_shape)
        log_var = extract_to(self.model.posterior_log_variance_clipped, ts, num_dim)
        return mean + noise_mask * (0.5 * log_var).exp() * noise


__all__ = [
    "BasicSampler",
]



import numpy as np

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Tuple
from typing import Optional

from .ddpm import make_condition_model
from .ddpm import DDPM
from ..ae.kl import GaussianDistribution
from ....zoo import DLZoo
from ....misc.toolkit import freeze
from ....misc.internal_ import register_custom_module


@register_custom_module("ldm")
class LDM(DDPM):
    def __init__(
        self,
        img_size: int,
        # unet
        in_channels: int,
        out_channels: int,
        *,
        num_heads: Optional[int] = 8,
        num_head_channels: Optional[int] = None,
        use_spatial_transformer: bool = False,
        num_transformer_layers: int = 1,
        context_dim: Optional[int] = None,
        signal_dim: int = 2,
        start_channels: int = 320,
        num_res_blocks: int = 2,
        attention_downsample_rates: Tuple[int, ...] = (1, 2, 4),
        dropout: float = 0.0,
        channel_multipliers: Tuple[int, ...] = (1, 2, 4, 4),
        resample_with_conv: bool = True,
        resample_with_resblock: bool = False,
        use_scale_shift_norm: bool = False,
        num_classes: Optional[int] = None,
        use_checkpoint: bool = False,
        attn_split_chunk: Optional[int] = None,
        # first stage
        first_stage: str,
        first_stage_config: Optional[Dict[str, Any]] = None,
        first_stage_scale_factor: float = 1.0,
        # diffusion
        ema_decay: Optional[float] = None,
        use_num_updates_in_ema: bool = True,
        parameterization: str = "eps",
        ## condition
        condition_type: str = "cross_attn",
        condition_model: Optional[str] = None,
        use_first_stage_as_condition: bool = False,
        condition_config: Optional[Dict[str, Any]] = None,
        condition_learnable: bool = False,
        ## noise schedule
        v_posterior: float = 0.0,
        timesteps: int = 1000,
        given_betas: Optional[np.ndarray] = None,
        beta_schedule: str = "linear",
        linear_start: float = 1.0e-4,
        linear_end: float = 2.0e-2,
        cosine_s: float = 8.0e-3,
        ## loss
        loss_type: str = "l2",
        l_simple_weight: float = 1.0,
        original_elbo_weight: float = 0.0,
        learn_log_var: bool = False,
        log_var_init: float = 0.0,
        ## sampling
        sampler: str = "ddim",
        sampler_config: Optional[Dict[str, Any]] = None,
    ):
        self.use_first_stage_as_condition = use_first_stage_as_condition
        if use_first_stage_as_condition:
            if condition_learnable:
                raise ValueError(
                    "should not use ae as condition model "
                    "when `condition_learnable` is set to True"
                )
            condition_model = None
        super().__init__(
            img_size,
            in_channels,
            out_channels,
            num_heads=num_heads,
            num_head_channels=num_head_channels,
            use_spatial_transformer=use_spatial_transformer,
            num_transformer_layers=num_transformer_layers,
            context_dim=context_dim,
            signal_dim=signal_dim,
            start_channels=start_channels,
            num_res_blocks=num_res_blocks,
            attention_downsample_rates=attention_downsample_rates,
            dropout=dropout,
            channel_multipliers=channel_multipliers,
            resample_with_conv=resample_with_conv,
            resample_with_resblock=resample_with_resblock,
            use_scale_shift_norm=use_scale_shift_norm,
            num_classes=num_classes,
            use_checkpoint=use_checkpoint,
            attn_split_chunk=attn_split_chunk,
            ema_decay=ema_decay,
            use_num_updates_in_ema=use_num_updates_in_ema,
            parameterization=parameterization,
            condition_type=condition_type,
            condition_model=condition_model,
            condition_config=condition_config,
            condition_learnable=condition_learnable,
            v_posterior=v_posterior,
            timesteps=timesteps,
            given_betas=given_betas,
            beta_schedule=beta_schedule,
            linear_start=linear_start,
            linear_end=linear_end,
            cosine_s=cosine_s,
            loss_type=loss_type,
            l_simple_weight=l_simple_weight,
            original_elbo_weight=original_elbo_weight,
            learn_log_var=learn_log_var,
            log_var_init=log_var_init,
            sampler=sampler,
            sampler_config=sampler_config,
        )
        first_stage_kw = first_stage_config or {}
        self.first_stage = freeze(DLZoo.load_model(first_stage, **first_stage_kw))
        self.scale_factor = first_stage_scale_factor
        # condition
        if use_first_stage_as_condition:
            # avoid recording duplicate parameters
            self.condition_model = [make_condition_model(first_stage, self.first_stage)]
        # sanity check
        embedding_channels = self.first_stage.core.embedding_channels
        if in_channels != embedding_channels and condition_type != "concat":
            raise ValueError(
                f"`in_channels` ({in_channels}) should be identical with the "
                f"`embedding_channels` ({embedding_channels}) of the "
                f"first_stage model ({first_stage})"
            )
        if out_channels != embedding_channels:
            raise ValueError(
                f"`out_channels` ({out_channels}) should be identical with the "
                f"`embedding_channels` ({embedding_channels}) of the "
                f"first_stage model ({first_stage})"
            )

    def decode(
        self,
        z: Tensor,
        *,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        start_step: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        latent = super().decode(
            z,
            cond=cond,
            num_steps=num_steps,
            start_step=start_step,
            verbose=verbose,
            **kwargs,
        )
        net = self._from_latent(latent)
        return net

    def _preprocess(self, net: Tensor, *, deterministic: bool = False) -> Tensor:
        net = self.first_stage.core.encode(net)
        if isinstance(net, GaussianDistribution):
            net = net.mode() if deterministic else net.sample()
        net = self.scale_factor * net
        return net

    def _from_latent(self, latent: Tensor) -> Tensor:
        latent = latent / self.scale_factor
        return self.first_stage.core.decode(latent, resize=False)

    def _get_cond(self, cond: Any) -> Tensor:
        if isinstance(self.condition_model, list):
            return self.condition_model[0](cond)
        return super()._get_cond(cond)


__all__ = [
    "LDM",
]



import torch

from torch import nn
from torch import Tensor
from typing import Optional
from cftool.types import tensor_dict_type

from .constants import LV1_ALPHA_KEY
from .constants import LV1_RAW_ALPHA_KEY
from ....constants import INPUT_KEY
from ....modules.blocks import get_conv_blocks
from ....modules.blocks import ResidualBlock
from ....misc.internal_.register import register_module


# TODO : Try ResidualBlockV2


@register_module("alpha_refine")
class AlphaRefineNet(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        *,
        num_layers: int = 3,
        latent_channels: int = 64,
        dropout: float = 0.0,
        ca_reduction: Optional[int] = None,
        eca_kernel_size: Optional[int] = None,
        lite: bool = False,
    ):
        super().__init__()
        if lite:
            latent_channels //= 2
        blocks = get_conv_blocks(
            in_channels + 1,
            latent_channels,
            3,
            1,
            activation=nn.ReLU(inplace=True),
        )
        for _ in range(num_layers - 2):
            if lite:
                blocks.extend(
                    get_conv_blocks(
                        latent_channels,
                        latent_channels,
                        3,
                        1,
                        activation=nn.ReLU6(inplace=True),
                    )
                )
            else:
                blocks.append(
                    ResidualBlock(
                        latent_channels,
                        dropout,
                        ca_reduction=ca_reduction,
                        eca_kernel_size=eca_kernel_size,
                    )
                )
        blocks.extend(get_conv_blocks(latent_channels, out_channels, 3, 1))
        self.net = nn.Sequential(*blocks)

    def forward(self, batch: tensor_dict_type) -> Tensor:
        net = batch[INPUT_KEY]
        lv1_raw_alpha = batch[LV1_RAW_ALPHA_KEY]
        lv1_alpha = batch[LV1_ALPHA_KEY]
        net = self.net(torch.cat([net, lv1_alpha], dim=1))
        net = net + lv1_raw_alpha
        return net

    def refine_from(self, net: Tensor, lv1_raw_alpha: Tensor) -> Tensor:
        alpha = torch.sigmoid(lv1_raw_alpha)
        inp = {INPUT_KEY: net, LV1_RAW_ALPHA_KEY: lv1_raw_alpha, LV1_ALPHA_KEY: alpha}
        return self.forward(inp)


__all__ = [
    "AlphaRefineNet",
]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import Type
from typing import Optional
from torch.nn import Module
from cftool.misc import shallow_copy_dict
from cftool.misc import WithRegister

from .norms import BN
from .customs import Linear
from .activations import Activation


mapping_dict: Dict[str, Type["MappingBase"]] = {}


class MappingBase(Module, WithRegister["MappingBase"]):
    d = mapping_dict

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__()


@MappingBase.register("basic")
class Mapping(MappingBase):
    def __init__(
        self,
        in_dim: int,
        out_dim: int,
        *,
        bias: Optional[bool] = None,
        pruner_config: Optional[dict] = None,
        dropout: float = 0.5,
        batch_norm: bool = True,
        activation: Optional[str] = "ReLU",
        init_method: str = "xavier_uniform",
        **kwargs: Any,
    ):
        super().__init__()
        self.config = shallow_copy_dict(kwargs)
        if bias is None:
            bias = not batch_norm
        self.linear = Linear(
            in_dim,
            out_dim,
            bias=bias,
            pruner_config=pruner_config,
            init_method=init_method,
            **shallow_copy_dict(kwargs),
        )
        self.bn = None if not batch_norm else BN(out_dim)
        if activation is None:
            self.activation: Optional[Module] = None
        else:
            activation_config = self.config.setdefault("activation_config", None)
            self.activation = Activation.make(activation, activation_config)
        use_dropout = 0.0 < dropout < 1.0
        self.dropout = None if not use_dropout else nn.Dropout(dropout)

    @property
    def weight(self) -> Tensor:
        return self.linear.weight

    @property
    def bias(self) -> Optional[Tensor]:
        return self.linear.bias

    def forward(self, net: Tensor) -> Tensor:
        net = self.linear(net)
        if self.bn is not None:
            net = self.bn(net)
        if self.activation is not None:
            net = self.activation(net)
        if self.dropout is not None:
            net = self.dropout(net)
        return net

    @classmethod
    def simple(
        cls,
        in_dim: int,
        out_dim: int,
        *,
        bias: bool = False,
        dropout: float = 0.0,
        batch_norm: bool = False,
        activation: Optional[str] = None,
        pruner_config: Optional[Dict[str, Any]] = None,
    ) -> "Mapping":
        if activation != "glu":
            activation_config = {}
        else:
            activation_config = {"in_dim": out_dim, "bias": bias}
        return cls(
            in_dim,
            out_dim,
            bias=bias,
            pruner_config=pruner_config,
            dropout=dropout,
            batch_norm=batch_norm,
            activation=activation,
            activation_config=activation_config,
        )


@MappingBase.register("res")
class ResBlock(MappingBase):
    to_latent: Module

    def __init__(
        self,
        in_dim: int,
        latent_dim: int,
        *,
        bias: Optional[bool] = None,
        pruner_config: Optional[dict] = None,
        dropout: float = 0.0,
        batch_norm: bool = True,
        activation: Optional[str] = "ReLU",
        init_method: str = "xavier_uniform",
        **kwargs: Any,
    ):
        super().__init__()
        # input mapping
        if in_dim == latent_dim:
            self.to_latent = nn.Identity()
        else:
            self.to_latent = Linear(
                in_dim,
                latent_dim,
                bias=True if bias is None else bias,
                pruner_config=pruner_config,
                init_method=init_method,
                **kwargs,
            )
        # residual unit
        self.residual_unit = nn.Sequential(
            BN(latent_dim),
            Activation.make(activation, kwargs.setdefault("activation_config", None)),
            nn.Identity() if not 0.0 < dropout < 1.0 else nn.Dropout(dropout),
            Mapping(
                latent_dim,
                latent_dim,
                bias=bias,
                pruner_config=pruner_config,
                dropout=dropout,
                batch_norm=batch_norm,
                activation=activation,
                init_method=init_method,
                **kwargs,
            ),
            Linear(
                latent_dim,
                latent_dim,
                bias=True if bias is None else bias,
                pruner_config=pruner_config,
                init_method=init_method,
                **kwargs,
            ),
        )

    def forward(self, net: Tensor) -> Tensor:
        net = self.to_latent(net)
        res = self.residual_unit(net)
        return net + res


@MappingBase.register("highway")
class HighwayBlock(MappingBase):
    def __init__(
        self,
        in_dim: int,
        latent_dim: int,
        *,
        bias: Optional[bool] = None,
        pruner_config: Optional[dict] = None,
        dropout: float = 0.0,
        batch_norm: bool = True,
        activation: Optional[str] = "ReLU",
        init_method: str = "xavier_uniform",
        **kwargs: Any,
    ):
        super().__init__()
        self.linear_mapping = Linear(
            in_dim,
            latent_dim,
            bias=True if bias is None else bias,
            pruner_config=pruner_config,
            init_method=init_method,
            **kwargs,
        )
        self.nonlinear_mapping = Mapping(
            in_dim,
            latent_dim,
            bias=bias,
            pruner_config=pruner_config,
            dropout=dropout,
            batch_norm=batch_norm,
            activation=activation,
            init_method=init_method,
            **kwargs,
        )
        self.gate_linear = Linear(
            in_dim,
            latent_dim,
            bias=True if bias is None else bias,
            pruner_config=pruner_config,
            init_method=init_method,
            **kwargs,
        )
        self.sigmoid = nn.Sigmoid()

    def forward(self, net: Tensor) -> Tensor:
        linear = self.linear_mapping(net)
        nonlinear = self.nonlinear_mapping(net)
        gate = self.sigmoid(self.gate_linear(net))
        return gate * nonlinear + (1.0 - gate) * linear


__all__ = [
    "mapping_dict",
    "MappingBase",
]



from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import Type
from cftool.misc import WithRegister


dl_zoo_model_loaders: Dict[str, Type["IDLZooModelLoader"]] = {}


class IDLZooModelLoader(WithRegister, metaclass=ABCMeta):
    d = dl_zoo_model_loaders

    @abstractmethod
    def permute_kwargs(self, kwargs: Dict[str, Any]) -> None:
        pass


__all__ = [
    "dl_zoo_model_loaders",
    "IDLZooModelLoader",
]



from typing import Optional

from ..register import register_ml_module
from ..protocols.ml import MixedStackedModel


@register_ml_module("fnet")
class FNet(MixedStackedModel):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        num_history: int,
        latent_dim: int = 256,
        *,
        num_layers: int = 4,
        dropout: float = 0.0,
        norm_type: Optional[str] = "batch_norm",
    ):
        super().__init__(
            input_dim,
            output_dim,
            num_history,
            latent_dim,
            token_mixing_type="fourier",
            num_layers=num_layers,
            dropout=dropout,
            norm_type=norm_type,
            feedforward_dim_ratio=1.0,
            use_head_token=False,
            use_positional_encoding=False,
        )


__all__ = ["FNet"]



import math
import torch

from typing import *
from torch import nn
from torch import Tensor
from torch.nn import init
from cftool.types import tensor_dict_type

from .protocol import Decoder1DMixin
from ....constants import INPUT_KEY
from ....constants import LABEL_KEY
from ....constants import LATENT_KEY
from ....misc.toolkit import auto_num_layers
from ....modules.blocks import Conv2d
from ....modules.blocks import Linear
from ....modules.blocks import GaussianBlur3
from ....modules.blocks import Activation
from ....modules.blocks import NormFactory
from ....modules.blocks import UpsampleConv2d


def style_mod(net: Tensor, style: Tensor) -> Tensor:
    style = style.view(style.shape[0], 2, net.shape[1], 1, 1)
    s0, s1 = style[:, 0], style[:, 1]
    return s1 + net * (s0 + 1.0)


class DecodeBlock(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        latent_dim: int,
        has_first_conv: bool,
        demodulate: bool,
        upscale: bool,
        activation: nn.Module,
        layer: int = 0,
    ):
        super().__init__()
        self.layer = layer
        self.activation = activation
        self.conv_1 = None
        if not has_first_conv:
            if in_channels != out_channels:
                raise ValueError(
                    "`in_channels` should equal to `out_channels` "
                    "when `has_first_conv` is False"
                )
        else:
            if not upscale:
                self.conv_1 = Conv2d(
                    in_channels,
                    out_channels,
                    kernel_size=3,
                    padding=1,
                    bias=False,
                    demodulate=demodulate,
                )
            else:
                self.conv_1 = UpsampleConv2d(
                    in_channels,
                    out_channels,
                    kernel_size=3,
                    transform_kernel=True,
                    padding="reflection",
                    bias=False,
                    demodulate=demodulate,
                    factor=2,
                )

        norm_factory = NormFactory("instance")
        norm_kwargs = {"affine": False, "eps": 1e-8}

        self.blur = GaussianBlur3(out_channels)
        self.noise_weight_1 = nn.Parameter(torch.empty(1, out_channels, 1, 1))
        self.noise_weight_1.data.zero_()
        self.bias_1 = nn.Parameter(torch.empty(1, out_channels, 1, 1))
        if demodulate:
            self.instance_norm_1 = None
        else:
            self.instance_norm_1 = norm_factory.make(out_channels, **norm_kwargs)
        self.style_1 = Linear(latent_dim, 2 * out_channels)

        self.conv_2 = Conv2d(
            out_channels,
            out_channels,
            kernel_size=3,
            padding=1,
            bias=False,
            demodulate=demodulate,
        )
        self.noise_weight_2 = nn.Parameter(torch.empty(1, out_channels, 1, 1))
        self.noise_weight_2.data.zero_()
        self.bias_2 = nn.Parameter(torch.empty(1, out_channels, 1, 1))
        if demodulate:
            self.instance_norm_2 = None
        else:
            self.instance_norm_2 = norm_factory.make(out_channels, **norm_kwargs)
        self.style_2 = Linear(latent_dim, 2 * out_channels)

        with torch.no_grad():
            self.bias_1.zero_()
            self.bias_2.zero_()

    def noise(self, net: Tensor, weight: Tensor, use_noise: bool) -> Tensor:
        if not use_noise:
            s = math.pow(self.layer + 1, 0.5)
            s = s * torch.exp(-net * net / (2.0 * s * s)) / math.sqrt(2 * math.pi) * 0.8
            net = net + s
        else:
            if use_noise == "batch_constant":
                shape = 1, 1, net.shape[2], net.shape[3]

            else:
                shape = net.shape[0], 1, net.shape[2], net.shape[3]
            noise_tensor = torch.randn(*shape)
            net = net + weight * noise_tensor.to(weight.device)
        return net

    def forward(
        self,
        net: Tensor,
        s1: Tensor,
        s2: Tensor,
        user_noise: bool = False,
    ) -> Tensor:
        if self.conv_1 is not None:
            net = self.conv_1(net)
            net = self.blur(net)

        net = self.noise(net, self.noise_weight_1, user_noise)
        net = net + self.bias_1
        if self.instance_norm_1 is not None:
            net = self.instance_norm_1(net)
        net = self.activation(net)
        net = style_mod(net, self.style_1(s1))
        net = self.conv_2(net)

        net = self.noise(net, self.noise_weight_2, user_noise)
        net = net + self.bias_2
        if self.instance_norm_2 is not None:
            net = self.instance_norm_2(net)
        net = self.activation(net)
        net = style_mod(net, self.style_2(s2))

        return net


class ToOutput(nn.Module):
    def __init__(
        self,
        in_channels: int,
        out_channels: int,
        gain: float = math.sqrt(2.0),
    ):
        super().__init__()
        self.mapping = Conv2d(in_channels, out_channels, kernel_size=1, gain=gain)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.mapping(x)


@Decoder1DMixin.register("style")
class StyleDecoder(nn.Module, Decoder1DMixin):
    def __init__(
        self,
        img_size: int,
        latent_dim: int = 128,
        out_channels: int = 3,
        *,
        start_channels: int = 64,
        latent_resolution: int = 4,
        max_channels: int = 256,
        num_classes: Optional[int] = None,
        noise: bool = False,
        demodulate: bool = False,
        num_upsample: Optional[int] = None,
    ):
        super().__init__()
        self.latent_dim = latent_dim
        target_num_upsample = auto_num_layers(img_size, latent_resolution, None) + 1
        if num_upsample is not None and num_upsample != target_num_upsample:
            raise ValueError(
                f"`num_upsample` is provided ({num_upsample}), "
                f"but it should be {target_num_upsample}"
            )
        num_upsample = target_num_upsample
        self._initialize(
            out_channels=out_channels,
            img_size=img_size,
            num_upsample=num_upsample,
            num_classes=num_classes,
            latent_resolution=latent_resolution,
        )
        self.noise = noise
        self.demodulate = demodulate
        self.start_channels = start_channels

        mul = 2 ** (num_upsample - 1)
        in_nc = min(max_channels, start_channels * mul)
        num_type = 1 if num_classes is None else num_classes
        resolution = latent_resolution
        self.const = nn.Parameter(torch.empty(num_type, in_nc, resolution, resolution))
        init.ones_(self.const.data)

        blocks: List[nn.Module] = []
        resolution //= 2
        for i in range(num_upsample):
            has_first_conv = i != 0
            upscale = resolution < img_size
            out_nc = min(max_channels, start_channels * mul)
            activation = Activation.make("leaky_relu_0.2" if upscale else "relu")
            blocks.append(
                DecodeBlock(
                    in_nc,
                    out_nc,
                    self.latent_dim,
                    has_first_conv,
                    self.demodulate,
                    upscale,
                    activation,
                    layer=i,
                )
            )
            resolution *= 2
            in_nc = out_nc
            mul //= 2

        self.blocks = nn.ModuleList(blocks)
        self.to_output = ToOutput(in_nc, out_channels, gain=0.03)

    def forward(self, batch: tensor_dict_type) -> Tensor:
        styles = batch.get(INPUT_KEY)
        if styles is None:
            msg = "`styles` should be provided in batch.data in `StyleGenerator`"
            raise ValueError(msg)

        labels = batch.get(LABEL_KEY)
        condition = batch.get(LATENT_KEY)
        if condition is not None:
            net = condition
        elif not self.is_conditional:
            net = self.const
        else:
            if labels is None:
                msg = "Conditional StyleDecoder is defined but labels are not provided"
                raise ValueError(msg)
            net = self.const[labels.view(-1)]

        for i in range(self.num_upsample):
            net = self.blocks[i](net, styles, styles, self.noise)  # type: ignore
        net = self.to_output(net)
        return net


__all__ = [
    "StyleDecoder",
]



import torch
import random

import numpy as np
import torch.nn as nn
import torch.nn.functional as F

from PIL import Image
from tqdm import tqdm
from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from cftool.misc import safe_execute
from cftool.misc import shallow_copy_dict
from cftool.array import save_images
from cftool.types import tensor_dict_type

from .common import read_image
from .common import restrict_wh
from .common import get_suitable_size
from .common import APIMixin
from ....zoo import DLZoo
from ....data import predict_tensor_data
from ....data import TensorInferenceData
from ....pipeline import DLPipeline
from ....constants import INPUT_KEY
from ....misc.toolkit import slerp
from ....misc.toolkit import new_seed
from ....misc.toolkit import eval_context
from ....misc.toolkit import seed_everything
from ....modules.blocks import Conv2d
from ....models.cv.diffusion import LDM
from ....models.cv.diffusion import DDPM
from ....models.cv.diffusion import ISampler
from ....models.cv.ae.common import IAutoEncoder
from ....models.cv.diffusion.utils import get_timesteps
from ....models.cv.diffusion.samplers.ddim import DDIMMixin
from ....models.cv.diffusion.samplers.k_samplers import KSamplerMixin


class switch_sampler_context:
    def __init__(self, api: "DiffusionAPI", sampler: str):
        self.api = api
        self.m_sampler = api.sampler.__identifier__
        self.target_sampler = sampler
        self.switched = False

    def __enter__(self) -> None:
        if self.m_sampler != self.target_sampler:
            self.switched = True
            self.api.switch_sampler(self.target_sampler)

    def __exit__(self, exc_type: Any, exc_val: Any, exc_tb: Any) -> None:
        if self.switched:
            self.api.switch_sampler(self.m_sampler)


class DiffusionAPI(APIMixin):
    m: DDPM
    sampler: ISampler
    cond_model: Optional[nn.Module]
    first_stage: Optional[IAutoEncoder]
    latest_seed: int
    latest_variation_seed: Optional[int]

    def __init__(
        self,
        m: DDPM,
        device: torch.device,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ):
        super().__init__(m, device, use_amp=use_amp, use_half=use_half)
        self.sampler = m.sampler
        self.cond_type = m.condition_type
        # extracted the condition model so we can pre-calculate the conditions
        self.cond_model = m.condition_model
        if self.cond_model is not None:
            self.cond_model.eval()
        m.condition_model = nn.Identity()
        # pre-calculate unconditional_cond if needed
        unconditional_cond = getattr(m.sampler, "unconditional_cond", None)
        if self.cond_model is not None and unconditional_cond is not None:
            uncond = self.get_cond(unconditional_cond)
            m.sampler.unconditional_cond = uncond.to(self.device)
        # extract first stage
        if not isinstance(m, LDM):
            self.first_stage = None
        else:
            self.first_stage = m.first_stage.core

    def get_cond(self, cond: Any) -> Tensor:
        if self.cond_model is None:
            msg = "should not call `get_cond` when `cond_model` is not available"
            raise ValueError(msg)
        with torch.no_grad():
            with self.amp_context:
                return self.cond_model(cond)

    def switch_sampler(
        self,
        sampler: str,
        sampler_config: Optional[Dict[str, Any]] = None,
    ) -> None:
        sampler_ins = self.m.make_sampler(sampler, sampler_config)
        current_unconditional_cond = getattr(self.m.sampler, "unconditional_cond", None)
        if current_unconditional_cond is not None:
            if hasattr(sampler_ins, "unconditional_cond"):
                sampler_ins.unconditional_cond = current_unconditional_cond
        current_guidance = getattr(self.m.sampler, "unconditional_guidance_scale", None)
        if current_guidance is not None:
            if hasattr(sampler_ins, "unconditional_guidance_scale"):
                sampler_ins.unconditional_guidance_scale = current_guidance
        self.sampler = self.m.sampler = sampler_ins

    def switch_circular(self, enable: bool) -> None:
        def _inject(m: nn.Module) -> None:
            for child in m.children():
                _inject(child)
            modules.append(m)

        padding_mode = "circular" if enable else "zeros"
        modules: List[nn.Module] = []
        _inject(self.m)
        for module in modules:
            if isinstance(module, nn.Conv2d):
                module.padding_mode = padding_mode
            elif isinstance(module, Conv2d):
                module.padding = padding_mode

    def sample(
        self,
        num_samples: int,
        export_path: Optional[str] = None,
        *,
        seed: Optional[int] = None,
        # each variation contains (seed, weight)
        variations: Optional[List[Tuple[int, float]]] = None,
        variation_seed: Optional[int] = None,
        variation_strength: Optional[float] = None,
        z: Optional[Tensor] = None,
        z_ref: Optional[Tensor] = None,
        z_ref_mask: Optional[Tensor] = None,
        z_ref_noise: Optional[Tensor] = None,
        size: Optional[Tuple[int, int]] = None,
        original_size: Optional[Tuple[int, int]] = None,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        unconditional_cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        batch_size: int = 1,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if cond is not None:
            if self.cond_type != "concat" and self.cond_model is not None:
                data = TensorInferenceData(cond, batch_size=batch_size)
                cond = predict_tensor_data(self.cond_model, data)
        if cond is not None and num_samples != len(cond):
            raise ValueError(
                f"`num_samples` ({num_samples}) should be identical with "
                f"the number of `cond` ({len(cond)})"
            )
        if alpha is not None and original_size is not None:
            alpha_h, alpha_w = alpha.shape[-2:]
            if alpha_w != original_size[0] or alpha_h != original_size[1]:
                raise ValueError(
                    f"shape of the provided `alpha` ({alpha_w}, {alpha_h}) should be "
                    f"identical with the provided `original_size` {original_size}"
                )
        unconditional = cond is None
        if unconditional:
            cond = [0] * num_samples
        iterator = TensorInferenceData(cond, batch_size=batch_size).initialize()[0]
        num_iter = len(iterator)
        if verbose and num_iter > 1:
            iterator = tqdm(iterator, desc="iter", total=num_iter)
        sampled = []
        kw = dict(num_steps=num_steps, verbose=verbose)
        kw.update(shallow_copy_dict(kwargs))
        if size is None:
            size = self.m.img_size, self.m.img_size
        else:
            if self.first_stage is None:
                factor = 1
            else:
                factor = self.first_stage.img_size // self.m.img_size
            size = tuple(map(lambda n: round(n / factor), size))  # type: ignore
        uncond_backup = None
        if self.cond_model is not None and unconditional_cond is not None:
            uncond_backup = self.sampler.unconditional_cond
            uncond = self.get_cond(unconditional_cond)
            self.sampler.unconditional_cond = uncond.to(self.device)
        with eval_context(self.m):
            with self.amp_context:
                for batch in iterator:
                    i_kw = shallow_copy_dict(kw)
                    i_cond = batch[INPUT_KEY].to(self.device)
                    repeat = lambda t: t.repeat_interleave(len(i_cond), dim=0)
                    if z is not None:
                        i_z = repeat(z)
                    else:
                        i_z_shape = len(i_cond), self.m.in_channels, *size[::-1]
                        i_z, _ = self._set_seed_and_variations(
                            seed,
                            lambda: torch.randn(i_z_shape, device=self.device),
                            lambda noise: noise,
                            variations,
                            variation_seed,
                            variation_strength,
                        )
                    if z_ref is not None and z_ref_mask is not None:
                        if z_ref_noise is not None:
                            i_kw["ref"] = repeat(z_ref)
                            i_kw["ref_mask"] = repeat(z_ref_mask)
                            i_kw["ref_noise"] = repeat(z_ref_noise)
                    if unconditional:
                        i_cond = None
                    if self.use_half:
                        i_z = i_z.half()
                        if i_cond is not None:
                            i_cond = i_cond.half()
                        for k, v in i_kw.items():
                            if isinstance(v, torch.Tensor) and v.is_floating_point():
                                i_kw[k] = v.half()
                    i_sampled = self.m.decode(i_z, cond=i_cond, **i_kw)
                    sampled.append(i_sampled.cpu().float())
        if uncond_backup is not None:
            self.sampler.unconditional_cond = uncond_backup
        concat = torch.cat(sampled, dim=0)
        if clip_output:
            concat = torch.clip(concat, -1.0, 1.0)
        if callback is not None:
            concat = callback(concat)
        if original_size is not None:
            with torch.no_grad():
                concat = F.interpolate(
                    concat,
                    original_size[::-1],
                    mode="bicubic",
                )
        if alpha is not None:
            alpha = torch.from_numpy(2.0 * alpha - 1.0)
            if original_size is None:
                with torch.no_grad():
                    alpha = F.interpolate(
                        alpha,
                        concat.shape[-2:],
                        mode="nearest",
                    )
            concat = torch.cat([concat, alpha], dim=1)
        if export_path is not None:
            save_images(concat, export_path)
        torch.cuda.empty_cache()
        return concat

    def txt2img(
        self,
        txt: Union[str, List[str]],
        export_path: Optional[str] = None,
        *,
        anchor: int = 64,
        max_wh: int = 512,
        num_samples: Optional[int] = None,
        size: Optional[Tuple[int, int]] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        callback: Optional[Callable[[Tensor], Tensor]] = None,
        batch_size: int = 1,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        input_is_str = isinstance(txt, str)
        if num_samples is None:
            num_samples = 1 if input_is_str else len(txt)
        if isinstance(txt, str):
            txt = [txt] * num_samples
        if len(txt) != num_samples:
            raise ValueError(
                f"`num_samples` ({num_samples}) should be identical with "
                f"the number of `txt` ({len(txt)})"
            )
        if size is None:
            new_size = None
        else:
            new_size = restrict_wh(*size, max_wh)
            new_size = tuple(map(get_suitable_size, new_size, (anchor, anchor)))  # type: ignore
        return self.sample(
            num_samples,
            export_path,
            size=new_size,
            original_size=size,
            cond=txt,
            num_steps=num_steps,
            clip_output=clip_output,
            callback=callback,
            batch_size=batch_size,
            verbose=verbose,
            **kwargs,
        )

    def outpainting(
        self,
        txt: str,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        fidelity: float = 0.2,
        padding_mode: str = "cv2_telea",
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        with switch_sampler_context(self, "plms"):
            res = read_image(image, max_wh, anchor=anchor, padding_mode=padding_mode)
            if res.alpha is None:
                raise ValueError("`image` should contain alpha channel in outpainting")
            z = z_ref = self._get_z(res.image)
            z_ref_mask = F.interpolate(
                torch.from_numpy(res.alpha).to(z_ref),
                z_ref.shape[-2:],
                mode="bicubic",
            )
            return self._img2img(
                z,
                export_path,
                z_ref=z_ref,
                z_ref_mask=z_ref_mask,
                fidelity=fidelity,
                original_size=res.original_size,
                alpha=None,
                cond=[txt],
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )

    def img2img(
        self,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        fidelity: float = 0.2,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        res = read_image(image, max_wh, anchor=anchor)
        z = self._get_z(res.image)
        return self._img2img(
            z,
            export_path,
            fidelity=fidelity,
            original_size=res.original_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def inpainting(
        self,
        image: Union[str, Image.Image],
        mask: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 32,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        refine_fidelity: Optional[float] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        # inpainting callback, will not trigger in refine stage
        def callback(out: Tensor) -> Tensor:
            final = torch.from_numpy(remained_image.copy())
            final += 0.5 * (1.0 + out) * (1.0 - remained_mask)
            return 2.0 * final - 1.0

        # handle mask stuffs
        image_res = read_image(image, max_wh, anchor=anchor)
        mask = read_image(mask, max_wh, anchor=anchor, to_mask=True).image
        bool_mask = mask >= 0.5
        remained_mask = (~bool_mask).astype(np.float16 if self.use_half else np.float32)
        remained_image = remained_mask * image_res.image
        # construct condition tensor
        remained_cond = self._get_z(remained_image)
        latent_shape = remained_cond.shape[-2:]
        mask_cond = torch.where(torch.from_numpy(bool_mask), 1.0, -1.0)
        mask_cond = mask_cond.to(torch.float16 if self.use_half else torch.float32)
        mask_cond = mask_cond.to(self.device)
        mask_cond = F.interpolate(mask_cond, size=latent_shape)
        cond = torch.cat([remained_cond, mask_cond], dim=1)
        # refine with img2img
        if refine_fidelity is not None:
            z = self._get_z(image_res.image)
            return self._img2img(
                z,
                export_path,
                fidelity=refine_fidelity,
                original_size=image_res.original_size,
                alpha=image_res.alpha if alpha is None else alpha,
                cond=cond,
                num_steps=num_steps,
                clip_output=clip_output,
                verbose=verbose,
                **kwargs,
            )
        # sampling
        z = torch.randn_like(remained_cond)
        return self.sample(
            1,
            export_path,
            z=z,
            original_size=image_res.original_size,
            alpha=image_res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            callback=callback,
            verbose=verbose,
            **kwargs,
        )

    def sr(
        self,
        image: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 8,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        if not isinstance(self.m, LDM):
            raise ValueError("`sr` is now only available for `LDM` models")
        factor = 2 ** (len(self.m.first_stage.core.channel_multipliers) - 1)
        res = read_image(image, round(max_wh / factor), anchor=anchor)
        wh_ratio = res.original_size[0] / res.original_size[1]
        zh, zw = res.image.shape[-2:]
        sr_size = (zw, zw / wh_ratio) if zw > zh else (zh * wh_ratio, zh)
        sr_size = tuple(map(lambda n: round(factor * n), sr_size))  # type: ignore
        cond = torch.from_numpy(2.0 * res.image - 1.0).to(self.device)
        z = torch.randn_like(cond)
        return self.sample(
            1,
            export_path,
            z=z,
            original_size=sr_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    def semantic2img(
        self,
        semantic: Union[str, Image.Image],
        export_path: Optional[str] = None,
        *,
        anchor: int = 16,
        max_wh: int = 512,
        alpha: Optional[np.ndarray] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        err_fmt = "`{}` is needed for `semantic2img`"
        if self.cond_model is None:
            raise ValueError(err_fmt.format("cond_model"))
        in_channels = getattr(self.cond_model, "in_channels", None)
        if in_channels is None:
            raise ValueError(err_fmt.format("cond_model.in_channels"))
        factor = getattr(self.cond_model, "factor", None)
        if factor is None:
            raise ValueError(err_fmt.format("cond_model.factor"))
        res = read_image(
            semantic,
            max_wh,
            anchor=anchor,
            to_gray=True,
            resample=Image.NEAREST,
            normalize=False,
        )
        cond = torch.from_numpy(res.image).to(torch.long).to(self.device)
        cond = F.one_hot(cond, num_classes=in_channels)[0]
        cond = cond.half() if self.use_half else cond.float()
        cond = cond.permute(0, 3, 1, 2).contiguous()
        cond = self.get_cond(cond)
        z = torch.randn_like(cond)
        return self.sample(
            1,
            export_path,
            z=z,
            original_size=res.original_size,
            alpha=res.alpha if alpha is None else alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )

    @classmethod
    def from_sd(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        return cls.from_pipeline(ldm_sd(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sd_anime(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_sd_anime()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_celeba_hq(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_celeba_hq()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_inpainting(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_inpainting()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_sr(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        return cls.from_pipeline(ldm_sr(), device, use_amp=use_amp, use_half=use_half)

    @classmethod
    def from_semantic(
        cls,
        device: Optional[str] = None,
        *,
        use_amp: bool = False,
        use_half: bool = False,
    ) -> "DiffusionAPI":
        m = ldm_semantic()
        return cls.from_pipeline(m, device, use_amp=use_amp, use_half=use_half)

    def _get_z(self, img: np.ndarray) -> Tensor:
        img = 2.0 * img - 1.0
        z = torch.from_numpy(img)
        if self.use_half:
            z = z.half()
        z = z.to(self.device)
        z = self.m._preprocess(z, deterministic=True)
        return z

    def _set_seed_and_variations(
        self,
        seed: Optional[int],
        get_noise: Callable[[], Tensor],
        get_new_z: Callable[[Tensor], Tensor],
        variations: Optional[List[Tuple[int, float]]],
        variation_seed: Optional[int],
        variation_strength: Optional[float],
    ) -> Tuple[Tensor, Tensor]:
        if seed is None:
            seed = new_seed()
        seed = seed_everything(seed)
        self.latest_seed = seed
        z_noise = get_noise()
        self.latest_variation_seed = None
        if variations is not None:
            for v_seed, v_weight in variations:
                seed_everything(v_seed)
                v_noise = get_noise()
                z_noise = slerp(v_noise, z_noise, v_weight)
        if variation_strength is not None:
            random.seed()
            if variation_seed is None:
                variation_seed = new_seed()
            variation_seed = seed_everything(variation_seed)
            self.latest_variation_seed = variation_seed
            variation_noise = get_noise()
            z_noise = slerp(variation_noise, z_noise, variation_strength)
        z = get_new_z(z_noise)
        return z, z_noise

    def _img2img(
        self,
        z: Tensor,
        export_path: Optional[str] = None,
        *,
        seed: Optional[int] = None,
        # each variation contains (seed, weight)
        variations: Optional[List[Tuple[int, float]]] = None,
        variation_seed: Optional[int] = None,
        variation_strength: Optional[float] = None,
        z_ref: Optional[Tensor] = None,
        z_ref_mask: Optional[Tensor] = None,
        fidelity: float = 0.2,
        original_size: Optional[Tuple[int, int]] = None,
        alpha: Optional[np.ndarray] = None,
        cond: Optional[Any] = None,
        num_steps: Optional[int] = None,
        clip_output: bool = True,
        verbose: bool = True,
        **kwargs: Any,
    ) -> Tensor:
        # perturb z
        if num_steps is None:
            num_steps = self.sampler.default_steps
        t = round((1.0 - fidelity) * num_steps)
        ts = get_timesteps(min(t, num_steps - 1), 1, z.device)
        if isinstance(self.sampler, (DDIMMixin, KSamplerMixin)):
            kw = shallow_copy_dict(self.sampler.sample_kwargs)
            kw["total_step"] = num_steps
            safe_execute(self.sampler._reset_buffers, kw)
        z, noise = self._set_seed_and_variations(
            seed,
            lambda: torch.randn_like(z),
            lambda noise_: self.sampler.q_sample(z, ts, noise_),
            variations,
            variation_seed,
            variation_strength,
        )
        kwargs["start_step"] = num_steps - t
        return self.sample(
            1,
            export_path,
            z=z,
            z_ref=z_ref,
            z_ref_mask=z_ref_mask,
            z_ref_noise=None if z_ref is None else noise,
            original_size=original_size,
            alpha=alpha,
            cond=cond,
            num_steps=num_steps,
            clip_output=clip_output,
            verbose=verbose,
            **kwargs,
        )


def _ldm(
    model: str,
    latent_size: int,
    latent_in_channels: int,
    latent_out_channels: int,
    **kwargs: Any,
) -> DLPipeline:
    kwargs["img_size"] = latent_size
    kwargs["in_channels"] = latent_in_channels
    kwargs["out_channels"] = latent_out_channels
    model_config = kwargs.setdefault("model_config", {})
    first_stage_kw = model_config.setdefault("first_stage_config", {})
    first_stage_kw.setdefault("report", False)
    first_stage_kw.setdefault("pretrained", True)
    first_stage_model_config = first_stage_kw.setdefault("model_config", {})
    use_loss = first_stage_model_config.setdefault("use_loss", False)
    if not use_loss:

        def state_callback(states: tensor_dict_type) -> tensor_dict_type:
            for key in list(states.keys()):
                if key.startswith("core.loss"):
                    states.pop(key)
            return states

        first_stage_kw["pretrained_state_callback"] = state_callback
    return DLZoo.load_pipeline(model, **kwargs)


def ldm(
    latent_size: int = 32,
    latent_in_channels: int = 4,
    latent_out_channels: int = 4,
    **kwargs: Any,
) -> DLPipeline:
    return _ldm(
        "diffusion/ldm",
        latent_size,
        latent_in_channels,
        latent_out_channels,
        **kwargs,
    )


def ldm_vq(
    latent_size: int = 64,
    latent_in_channels: int = 3,
    latent_out_channels: int = 3,
    **kwargs: Any,
) -> DLPipeline:
    return _ldm(
        "diffusion/ldm.vq",
        latent_size,
        latent_in_channels,
        latent_out_channels,
        **kwargs,
    )


def ldm_sd(pretrained: bool = True, **kwargs: Any) -> DLPipeline:
    return _ldm("diffusion/ldm.sd", 64, 4, 4, pretrained=pretrained, **kwargs)


def ldm_sd_anime(pretrained: bool = True) -> DLPipeline:
    return ldm_sd(pretrained, download_name="ldm_sd_anime_nai")


def ldm_celeba_hq(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        download_name="ldm_celeba_hq",
        model_config=dict(
            ema_decay=None,
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


def ldm_inpainting(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        latent_in_channels=7,
        download_name="ldm_inpainting",
        model_config=dict(
            ema_decay=None,
            start_channels=256,
            num_heads=8,
            num_head_channels=None,
            resample_with_resblock=True,
            condition_type="concat",
            first_stage_config=dict(
                pretrained=False,
                model_config=dict(
                    attention_type="none",
                ),
            ),
        ),
    )


def ldm_sr(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        latent_in_channels=6,
        download_name="ldm_sr",
        model_config=dict(
            ema_decay=None,
            start_channels=160,
            attention_downsample_rates=[8, 16],
            channel_multipliers=[1, 2, 2, 4],
            condition_type="concat",
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


def ldm_semantic(pretrained: bool = True) -> DLPipeline:
    return ldm_vq(
        pretrained=pretrained,
        latent_size=128,
        latent_in_channels=6,
        download_name="ldm_semantic",
        model_config=dict(
            ema_decay=None,
            start_channels=128,
            num_heads=8,
            num_head_channels=None,
            attention_downsample_rates=[8, 16, 32],
            channel_multipliers=[1, 4, 8],
            condition_type="concat",
            condition_model="rescaler",
            condition_config=dict(
                num_stages=2,
                in_channels=182,
                out_channels=3,
            ),
            first_stage_config=dict(
                pretrained=False,
            ),
        ),
    )


__all__ = [
    "DiffusionAPI",
]



import os
import json
import math
import torch
import shutil
import inspect
import zipfile

from abc import abstractmethod
from abc import ABCMeta
from typing import Any
from typing import Dict
from typing import List
from typing import Type
from typing import Tuple
from typing import Union
from typing import Callable
from typing import Optional
from typing import NamedTuple
from typing import OrderedDict as OrderedDictType
from collections import OrderedDict
from cftool.misc import random_hash
from cftool.misc import safe_execute
from cftool.misc import print_warning
from cftool.misc import check_requires
from cftool.misc import shallow_copy_dict
from cftool.misc import prepare_workplace_from
from cftool.misc import truncate_string_to_length
from cftool.misc import lock_manager
from cftool.misc import Saving
from cftool.misc import WithRegister
from cftool.types import np_dict_type
from cftool.types import tensor_dict_type

from .data import DataModule
from .data import DLDataModule
from .types import configs_type
from .types import sample_weights_type
from .types import states_callback_type
from .trainer import get_sorted_checkpoints
from .trainer import Trainer
from .protocol import loss_dict
from .protocol import callback_dict
from .protocol import DeviceInfo
from .protocol import ILoss
from .protocol import IDLModel
from .protocol import _IMetric
from .protocol import MetricsOutputs
from .protocol import InferenceOutputs
from .protocol import IInference
from .protocol import IDataLoader
from .protocol import ModelWithCustomSteps
from .constants import PT_PREFIX
from .constants import SCORES_FILE
from .constants import CHECKPOINTS_FOLDER
from .constants import BATCH_INDICES_KEY
from .misc.toolkit import get_ddp_info
from .misc.toolkit import _get_environ_workplace
from .misc.toolkit import ConfigMeta
from .misc.internal_.trainer import make_trainer


pipeline_dict: Dict[str, Type["IPipeline"]] = {}
dl_pipeline_modifiers: Dict[str, Type["IModifier"]] = {}


class IPipeline(WithRegister["IPipeline"], metaclass=ABCMeta):
    d = pipeline_dict

    @abstractmethod
    def build(self, data_info: Dict[str, Any]) -> None:
        pass

    @abstractmethod
    def fit(
        self,
        data: DataModule,
        *,
        sample_weights: sample_weights_type = None,
    ) -> "IPipeline":
        pass

    @abstractmethod
    def predict(self, data: DataModule, **predict_kwargs: Any) -> np_dict_type:
        pass

    @abstractmethod
    def save(self, export_folder: str) -> "IPipeline":
        pass

    @staticmethod
    @abstractmethod
    def load(export_folder: str) -> "IPipeline":
        pass


class ModelSoupConfigs(NamedTuple):
    loader: IDataLoader
    metric_names: Union[str, List[str]]
    metric_configs: configs_type = None
    metric_weights: Optional[Dict[str, float]] = None
    valid_portion: float = 1.0
    strategy: str = "greedy"
    states_callback: states_callback_type = None
    verbose: bool = True


def _generate_model_soup_checkpoint(
    m: "DLPipeline",
    checkpoint_folder: str,
    configs: ModelSoupConfigs,
    scores_export_path: str,
    checkpoint_export_path: str,
) -> None:
    current_states: tensor_dict_type = {}
    with open(os.path.join(checkpoint_folder, SCORES_FILE), "r") as f:
        scores = json.load(f)
    sorted_files = get_sorted_checkpoints(checkpoint_folder)
    sorted_paths = [os.path.join(checkpoint_folder, file) for file in sorted_files]
    metrics = _IMetric.fuse(
        configs.metric_names,
        configs.metric_configs,
        metric_weights=configs.metric_weights,
    )
    chosen_ingredients = []
    if configs.strategy == "greedy":
        n = 0
        best_score = -math.inf
        if configs.verbose:
            print("\n".join(["=" * 100, "Creating Model Soup", "-" * 100]))
        for file, path in zip(sorted_files, sorted_paths):
            if configs.verbose:
                print(f"> Checking {file}")
            states = torch.load(path, map_location=m.device)
            m._make_modifier().permute_states(states)
            states_backup = shallow_copy_dict(current_states)
            if configs.states_callback is not None:
                states = configs.states_callback(m, states)
            for k, v in states.items():
                if n == 0:
                    current_states[k] = v
                else:
                    current_v = current_states[k]
                    current_states[k] = (current_v * n + v) / (n + 1)
            m.model.load_state_dict(current_states)
            kw = dict(
                portion=configs.valid_portion,
                metrics=metrics,
            )
            if check_requires(m.inference.get_outputs, "use_loader_cache"):
                kw["use_loader_cache"] = False
            res = m.inference.get_outputs(configs.loader, **kw)  # type: ignore
            score = res.metric_outputs.final_score  # type: ignore
            if score < best_score:
                current_states = states_backup
                if configs.verbose:
                    print(">> Performance is not improving")
            else:
                n += 1
                best_score = score
                if configs.verbose:
                    chosen_ingredients.append(f"{file} ({scores[file]})")
                    print(f">> New SOTA! ({score})")
    else:
        msg = f"model soup strategy `{configs.strategy}` is not implemented yet"
        raise NotImplementedError(msg)
    with open(scores_export_path, "r") as rf:
        existing_scores = json.load(rf)
    key = list(existing_scores.keys())[0]
    with open(scores_export_path, "w") as wf:
        json.dump({key: best_score}, wf)
    torch.save(current_states, checkpoint_export_path)
    original_metric_names = m.trainer_config["metric_names"]
    if original_metric_names is None:
        metrics_identifier = "+ DEFAULTS +"
    else:
        if isinstance(original_metric_names, str):
            original_metric_names = [original_metric_names]
        metrics_identifier = " | ".join(original_metric_names)
    if configs.verbose:
        print(
            "\n".join(
                [
                    "=" * 100,
                    "Model Soup Generated",
                    "-" * 100,
                    f"Ingredients (Original Metrics : {metrics_identifier})",
                    "-" * 100,
                    "\n".join(chosen_ingredients),
                    "-" * 100,
                    "Final Score",
                    "-" * 100,
                    str(best_score),
                    "-" * 100,
                ]
            )
        )


class IDLPipeline:
    modifier: str

    _defaults: OrderedDictType[str, Any]

    data: Optional[DLDataModule]
    data_info: Dict[str, Any]
    data_module_bytes: Optional[bytes]
    loss: ILoss
    loss_name: str
    loss_config: Optional[Dict[str, Any]]
    model: IDLModel
    model_name: str
    model_config: Dict[str, Any]
    trainer: Trainer
    trainer_config: Dict[str, Any]
    inference: IInference
    inference_base: Type[IInference]
    device_info: DeviceInfo

    pipeline_key: str
    pipeline_file: str
    config_name: str
    trainer_config_file: str
    data_info_name: str
    metrics_log_file: str

    final_results_file: str
    config_bundle_name: str

    config: Dict[str, Any]
    input_dim: Optional[int]

    built: bool
    in_loading: bool
    need_build_trainer: bool

    device: torch.device
    is_rank_0: bool


def get_requirements(c: Type, *, excludes: List[str]) -> List[str]:
    lines = inspect.getsource(c).split("\n")[1:]
    requirements = list(filter(bool, (line.strip().split(":")[0] for line in lines)))
    for e in excludes:
        requirements.remove(e)
    return requirements


class IModifier(WithRegister["IModifier"], IDLPipeline):
    d = dl_pipeline_modifiers

    build_steps = [
        "record_num_samples",
        "prepare_workplace",
        "prepare_loss",
        "build_model",
        "build_inference",
    ]
    requirements = get_requirements(IDLPipeline, excludes=["data"])

    def __init__(self, pipeline: "DLPipeline") -> None:
        self.__pipeline = pipeline

    def __getattr__(self, __name: str) -> Any:
        return getattr(self.__pipeline, __name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        pipeline_key = "_IModifier__pipeline"
        if __name == pipeline_key:
            self.__dict__[pipeline_key] = __value
        else:
            setattr(self.__pipeline, __name, __value)

    def _write_pipeline_info(self, folder: str) -> None:
        with open(os.path.join(folder, self.pipeline_file), "w") as f:
            f.write(self.__pipeline.__identifier__)

    def _sanity_check(self) -> None:
        missing_requirements = []
        token = f"{random_hash()}{random_hash()}{random_hash()}"
        for requirement in self.requirements:
            if getattr(self.__pipeline, requirement, token) is token:
                missing_requirements.append(requirement)
        if missing_requirements:
            pipeline_name = self.__pipeline.__class__.__name__
            raise ValueError(
                f"the following attributes is missing in `{pipeline_name}`: "
                f"{', '.join(missing_requirements)}"
            )

    @staticmethod
    def _report_messages(title: str, messages: Dict[str, Any]) -> None:
        def _stringify_item(item: Tuple[str, Any], prefix: Optional[str] = None) -> str:
            key, value = item
            if prefix is not None:
                key = f"{prefix}{key}"
            if not isinstance(value, dict) or not value:
                key = truncate_string_to_length(key, span)
                return f"{key:>{span}s}   |   {value}"
            prefix = f"{key}."
            items = [_stringify_item((vk, vv), prefix) for vk, vv in value.items()]
            return "\n".join(items)

        span = 64
        length = 2 * span
        print(
            "\n".join(
                [
                    "=" * length,
                    f"{title:^{length}s}",
                    "-" * length,
                    "\n".join(map(_stringify_item, messages.items())),
                    "-" * length,
                ]
            )
        )

    def _report_defaults(self) -> None:
        self._report_messages(
            "Internal Default Configurations Used by `carefree-learn`",
            self._defaults,
        )

    def _report_configs(self) -> None:
        self._report_messages("External Configurations", self.config)

    # build steps

    def record_num_samples(self) -> None:
        if not self.need_build_trainer:
            return
        data: Optional[DLDataModule] = getattr(self, "data", None)
        if data is None:
            self._defaults["train_samples"] = None
            self._defaults["valid_samples"] = None
        else:
            self._defaults["train_samples"] = len(data.train_data)
            if data.valid_data is None:
                self._defaults["valid_samples"] = None
            else:
                self._defaults["valid_samples"] = len(data.valid_data)
        self._defaults.move_to_end("valid_samples", last=False)
        self._defaults.move_to_end("train_samples", last=False)

    def prepare_workplace(self) -> None:
        if self.is_rank_0 and not self.in_loading and self.need_build_trainer:
            workplace = prepare_workplace_from(self.trainer_config["workplace"])
            self._defaults["workplace"] = workplace
            self.trainer_config["workplace"] = workplace
            self.trainer_config["data_info_name"] = self.data_info_name
            self.trainer_config["metrics_log_file"] = self.metrics_log_file
            self._write_pipeline_info(workplace)
            Saving.save_dict(self.config, self.config_name, workplace)

    def prepare_loss(self) -> None:
        if self.in_loading:
            return None
        self.loss_name = ILoss.parse(self.loss_name)
        self.loss = ILoss.make(self.loss_name, self.loss_config or {})

    def build_model(self, data_info: Dict[str, Any]) -> None:
        self.model = IDLModel.make(self.model_name, config=self.model_config)

    def build_inference(self) -> None:
        self.inference = self.inference_base(model=self.model)

    def prepare_trainer_defaults(self, data_info: Dict[str, Any]) -> None:
        # set some trainer defaults to deep learning tasks which work well in practice
        if get_ddp_info() is not None:
            mns = self.trainer_config["monitor_names"]
            if mns is not None and mns != "conservative" and mns != ["conservative"]:
                print_warning(
                    "only `conservative` monitor is available "
                    f"in DDP mode, {mns} found"
                )
            self.trainer_config["monitor_names"] = "conservative"
        if self.trainer_config["monitor_names"] is None:
            self._defaults["monitor_names"] = "conservative"
            self.trainer_config["monitor_names"] = "conservative"
        tqdm_settings = self.trainer_config["tqdm_settings"]
        callback_names = self.trainer_config["callback_names"]
        callback_configs = self.trainer_config["callback_configs"]
        optimizer_settings = self.trainer_config["optimizer_settings"]
        if callback_names is None:
            callback_names = []
        if callback_configs is None:
            callback_configs = {}
        if isinstance(callback_names, str):
            callback_names = [callback_names]
        auto_callback = self.trainer_config.get("auto_callback", True)
        if "mlflow" in callback_names and auto_callback:
            mlflow_config = callback_configs.setdefault("mlflow", {})
            if "experiment_name" not in mlflow_config:
                mlflow_config["experiment_name"] = self.model.__identifier__
                self._defaults["mlflow_experiment_name"] = self.model.__identifier__
        if "_log_metrics_msg" not in callback_names and auto_callback:
            self._defaults["additional_callbacks"] = ["_log_metrics_msg"]
            callback_names.insert(0, "_log_metrics_msg")
            verbose = False
            if tqdm_settings is None or not tqdm_settings.get("use_tqdm", False):
                verbose = True
            log_metrics_msg_config = callback_configs.setdefault("_log_metrics_msg", {})
            if "verbose" not in log_metrics_msg_config:
                log_metrics_msg_config["verbose"] = verbose
                self._defaults["log_metrics_msg_verbose"] = verbose
        self.trainer_config["tqdm_settings"] = tqdm_settings
        self.trainer_config["callback_names"] = callback_names
        self.trainer_config["callback_configs"] = callback_configs
        self.trainer_config["optimizer_settings"] = optimizer_settings

    ## api

    def build(
        self,
        data_info: Dict[str, Any],
        *,
        build_trainer: bool = True,
        report: Optional[bool] = None,
    ) -> None:
        if self.built:
            if build_trainer and self.trainer is None:
                self.need_build_trainer = True
                self.record_num_samples()
                self.prepare_workplace()
                self.build_trainer()
                self.report()
            return None
        self.data_info = shallow_copy_dict(data_info)
        self.need_build_trainer = build_trainer
        kw = dict(data_info=data_info)
        for step in self.build_steps:
            safe_execute(getattr(self, step), shallow_copy_dict(kw))
        if self.in_loading:
            return None
        safe_execute(self.prepare_trainer_defaults, shallow_copy_dict(kw))
        if not build_trainer:
            self.trainer = None  # type: ignore
        else:
            self.build_trainer()
        self._sanity_check()
        if report is None:
            if self.trainer is None:
                report = False
            else:
                is_rank_0 = self.trainer.is_rank_0
                in_distributed = self.trainer.tqdm_settings.in_distributed
                report = is_rank_0 and not in_distributed
        if report:
            self.report()
        self.built = True

    def build_trainer(self) -> None:
        trainer_config = shallow_copy_dict(self.trainer_config)
        if isinstance(self.model, ModelWithCustomSteps):
            self.model.permute_trainer_config(trainer_config)
        self.trainer = make_trainer(**trainer_config)

    def report(self) -> None:
        self._report_defaults()
        self._report_configs()

    # load steps

    def load_states_from(self, folder: str) -> Dict[str, Any]:
        checkpoints = get_sorted_checkpoints(folder)
        if not checkpoints:
            raise ValueError(f"no model file found in {folder}")
        checkpoint_path = os.path.join(folder, checkpoints[0])
        return torch.load(checkpoint_path, map_location=self.device)

    # changes should happen inplace
    def permute_states(self, states: Dict[str, Any]) -> None:
        pass

    ## api

    def save_misc(self, export_folder: str) -> float:
        os.makedirs(export_folder, exist_ok=True)
        self._write_pipeline_info(export_folder)
        data: Optional[DLDataModule] = getattr(self, "data", None)
        if data is not None:
            data.save_info(export_folder)
        else:
            if self.data_module_bytes is None:
                print_warning(
                    "`data_module_bytes` is not found. This is likely to cause "
                    "data-processing issues unless you are "
                    "saving a pretrained model."
                )
            else:
                msg = "`data` is not found, `pipeline.data_module_bytes` will be saved"
                print_warning(msg)
                data_folder = os.path.join(export_folder, DLDataModule.package_folder)
                zip_path = f"{data_folder}.zip"
                with open(zip_path, "wb") as f:
                    f.write(self.data_module_bytes)
                with zipfile.ZipFile(zip_path, "r") as zip_ref:
                    zip_ref.extractall(data_folder)
                os.remove(zip_path)
        # final results
        final_results = None
        try:
            final_results = self.trainer.final_results
            if final_results is None:
                print_warning(
                    "`final_results` are not generated yet, "
                    "'unknown' results will be saved"
                )
        except AttributeError as e:
            print_warning(
                f"{e}, so `final_results` cannot be accessed, "
                "and 'unknown' results will be saved"
            )
        if final_results is None:
            final_results = MetricsOutputs(0.0, {"unknown": 0.0})
        with open(os.path.join(export_folder, self.final_results_file), "w") as f:
            json.dump(final_results, f)
        # config bundle
        config_bundle = {
            "config": shallow_copy_dict(self.config),
            "device_info": self.device_info,
        }
        Saving.save_dict(config_bundle, self.config_bundle_name, export_folder)
        return final_results.final_score

    @staticmethod
    def load_infrastructure(
        cls: Type["DLPipeline"],
        export_folder: str,
        cuda: Optional[str],
        to_original_device: bool,
        pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        post_callback: Optional[Callable[["DLPipeline", Dict[str, Any]], None]] = None,
    ) -> "DLPipeline":
        config_bundle = Saving.load_dict(cls.config_bundle_name, export_folder)
        if pre_callback is not None:
            pre_callback(config_bundle)
        config = config_bundle["config"]
        config["in_loading"] = True
        m = safe_execute(cls, config)
        device_info = DeviceInfo(*config_bundle["device_info"])
        if not to_original_device:
            device_info = device_info._replace(cuda=cuda)
        elif cuda is not None:
            print_warning(
                "`to_original_device` is set to True, so "
                f"`cuda={cuda}` will be ignored"
            )
        m.device_info = device_info
        if post_callback is not None:
            post_callback(m, config_bundle)
        return m

    # inference

    def postprocess(self, outputs: InferenceOutputs) -> np_dict_type:
        return outputs.forward_results


@IPipeline.register("dl")
class DLPipeline(IPipeline, IDLPipeline, metaclass=ConfigMeta):
    modifier = "dl"

    inference_base = IInference

    pipeline_key = "pipeline"
    pipeline_file = "pipeline.txt"
    config_name = "config"
    trainer_config_file = "trainer_config.json"
    data_info_name = "data_info"
    metrics_log_file = "metrics.txt"

    final_results_file = "final_results.json"
    config_bundle_name = "config_bundle"

    def __init__(
        self,
        model_name: str,
        model_config: Optional[Dict[str, Any]] = None,
        *,
        loss_name: Optional[str] = None,
        loss_config: Optional[Dict[str, Any]] = None,
        # trainer
        state_config: Optional[Dict[str, Any]] = None,
        num_epoch: int = 40,
        max_epoch: int = 1000,
        fixed_epoch: Optional[int] = None,
        fixed_steps: Optional[int] = None,
        log_steps: Optional[int] = None,
        valid_portion: float = 1.0,
        amp: bool = False,
        clip_norm: float = 0.0,
        cudnn_benchmark: bool = False,
        metric_names: Optional[Union[str, List[str]]] = None,
        metric_configs: configs_type = None,
        metric_weights: Optional[Dict[str, float]] = None,
        use_losses_as_metrics: Optional[bool] = None,
        loss_metrics_weights: Optional[Dict[str, float]] = None,
        recompute_train_losses_in_eval: bool = True,
        monitor_names: Optional[Union[str, List[str]]] = None,
        monitor_configs: Optional[Dict[str, Any]] = None,
        callback_names: Optional[Union[str, List[str]]] = None,
        callback_configs: Optional[Dict[str, Any]] = None,
        lr: Optional[float] = None,
        optimizer_name: Optional[str] = None,
        scheduler_name: Optional[str] = None,
        optimizer_config: Optional[Dict[str, Any]] = None,
        scheduler_config: Optional[Dict[str, Any]] = None,
        update_scheduler_per_epoch: bool = False,
        optimizer_settings: Optional[Dict[str, Dict[str, Any]]] = None,
        use_zero: bool = False,
        workplace: str = "_logs",
        finetune_config: Optional[Dict[str, Any]] = None,
        tqdm_settings: Optional[Dict[str, Any]] = None,
        # misc
        in_loading: bool = False,
        allow_no_loss: bool = False,
    ):
        # record defaults
        self._defaults = OrderedDict()
        # sanity check
        if loss_name is None:
            if model_name in loss_dict:
                loss_name = model_name
            else:
                model_base = IDLModel.get(model_name)
                if allow_no_loss or issubclass(model_base, ModelWithCustomSteps):
                    loss_name = ILoss.placeholder_key
                else:
                    raise ValueError(
                        "`loss_name` should be provided when "
                        f"`{model_name}` has not implemented its own loss "
                        "and `allow_no_loss` is False"
                    )
            self._defaults["loss_name"] = loss_name
        # set defaults
        if state_config is None:
            state_config = {}
        if "max_snapshot_file" not in state_config:
            state_config["max_snapshot_file"] = 25
            self._defaults["max_snapshot_file"] = 25
        if callback_names is None:
            if model_name in callback_dict:
                callback_names = model_name
                self._defaults["callback_names"] = model_name
        # initialize
        self.input_dim = None
        self.model_name = model_name
        self.model_config = model_config or {}
        self.data_module_bytes = None
        self.loss_name = loss_name
        self.loss_config = loss_config
        self.trainer_config: Dict[str, Any] = {
            "state_config": state_config,
            "num_epoch": num_epoch,
            "max_epoch": max_epoch,
            "fixed_epoch": fixed_epoch,
            "fixed_steps": fixed_steps,
            "log_steps": log_steps,
            "valid_portion": valid_portion,
            "amp": amp,
            "clip_norm": clip_norm,
            "metric_names": metric_names,
            "metric_configs": metric_configs,
            "metric_weights": metric_weights,
            "use_losses_as_metrics": use_losses_as_metrics,
            "loss_metrics_weights": loss_metrics_weights,
            "recompute_train_losses_in_eval": recompute_train_losses_in_eval,
            "monitor_names": monitor_names,
            "monitor_configs": monitor_configs,
            "callback_names": callback_names,
            "callback_configs": callback_configs,
            "lr": lr,
            "optimizer_name": optimizer_name,
            "scheduler_name": scheduler_name,
            "optimizer_config": optimizer_config,
            "scheduler_config": scheduler_config,
            "update_scheduler_per_epoch": update_scheduler_per_epoch,
            "optimizer_settings": optimizer_settings,
            "use_zero": use_zero,
            "workplace": _get_environ_workplace() or workplace,
            "finetune_config": finetune_config,
            "tqdm_settings": tqdm_settings,
        }
        self.in_loading = in_loading
        self.built = False
        self.device_info = DeviceInfo(None, None)
        if cudnn_benchmark:
            torch.backends.cudnn.benchmark = True

    def __eq__(self, __o: object) -> bool:
        if not isinstance(__o, DLPipeline):
            return False
        return json.dumps(self.to_json()) == json.dumps(__o.to_json())

    @property
    def device(self) -> torch.device:  # type: ignore
        return self.device_info.device

    @property
    def is_rank_0(self) -> bool:  # type: ignore
        ddp_info = get_ddp_info()
        if ddp_info is None:
            return True
        if ddp_info.rank == 0:
            return True
        return False

    def _make_modifier(self) -> IModifier:
        return IModifier.make(self.modifier, {"pipeline": self})

    # api

    def build(
        self,
        data_info: Dict[str, Any],
        *,
        build_trainer: bool = True,
        report: Optional[bool] = None,
    ) -> None:
        modifier = self._make_modifier()
        modifier.build(data_info, build_trainer=build_trainer, report=report)

    def fit(  # type: ignore
        self,
        data: DLDataModule,
        *,
        sample_weights: sample_weights_type = None,
        cuda: Optional[Union[int, str]] = None,
    ) -> "DLPipeline":
        data.prepare(sample_weights)
        if cuda is not None:
            cuda = str(cuda)
        self.data = data
        self.build(data.info)
        self.trainer.fit(
            data,
            self.loss,
            self.model,
            self.inference,
            config_export_file=self.trainer_config_file,
            cuda=cuda,
        )
        self.device_info = self.trainer.device_info
        return self

    def predict(self, data: DLDataModule, **predict_kwargs: Any) -> np_dict_type:
        train_loader, valid_loader = data.initialize()
        if valid_loader is not None:
            raise ValueError("`valid_loader` should not be provided in `predict`")
        kw0 = shallow_copy_dict(predict_kwargs)
        kw1 = shallow_copy_dict(predict_kwargs)
        kw0["loader"] = train_loader
        outputs = safe_execute(self.inference.get_outputs, kw0)
        kw1["outputs"] = outputs
        return safe_execute(self._make_modifier().postprocess, kw1)

    def save(
        self,
        export_folder: str,
        *,
        compress: bool = True,
        remove_original: bool = True,
    ) -> "DLPipeline":
        abs_folder = os.path.abspath(export_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [export_folder]):
            score = self._make_modifier().save_misc(export_folder)
            if getattr(self, "trainer", None) is not None:
                if getattr(self.trainer, "model", None) is None:
                    self.trainer.model = self.model
                self.trainer.save_checkpoint(score, export_folder, no_history=True)
            else:
                file = f"{PT_PREFIX}-1.pt"
                torch.save(self.model.state_dict(), os.path.join(export_folder, file))
                with open(os.path.join(export_folder, SCORES_FILE), "w") as f:
                    json.dump({file: 0.0}, f)
            if compress:
                Saving.compress(abs_folder, remove_original=remove_original)
        return self

    @classmethod
    def pack(
        cls,
        workplace: str,
        *,
        step: Optional[str] = None,
        config_bundle_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
        pack_folder: Optional[str] = None,
        cuda: Optional[Union[int, str]] = None,
        compress: bool = True,
        model_soup_configs: Optional[ModelSoupConfigs] = None,
    ) -> str:
        if cuda is not None:
            cuda = str(cuda)
        if pack_folder is None:
            pack_name = f"packed{'' if step is None else f'_{step}'}"
            pack_folder = os.path.join(workplace, pack_name)
        if os.path.isdir(pack_folder):
            print_warning(f"'{pack_folder}' already exists, it will be erased")
            shutil.rmtree(pack_folder)
        os.makedirs(pack_folder)
        abs_folder = os.path.abspath(pack_folder)
        base_folder = os.path.dirname(abs_folder)
        with lock_manager(base_folder, [pack_folder]):
            shutil.copyfile(
                os.path.join(workplace, cls.pipeline_file),
                os.path.join(pack_folder, cls.pipeline_file),
            )
            checkpoint_folder = os.path.join(workplace, CHECKPOINTS_FOLDER)
            if step is not None:
                best_file = f"{PT_PREFIX}{step}.pt"
            else:
                best_file = get_sorted_checkpoints(checkpoint_folder)[0]
            new_file = f"{PT_PREFIX}-1.pt"
            new_path = os.path.join(pack_folder, new_file)
            shutil.copyfile(os.path.join(checkpoint_folder, best_file), new_path)
            with open(os.path.join(checkpoint_folder, SCORES_FILE), "r") as rf:
                scores = json.load(rf)
            new_scores_path = os.path.join(pack_folder, SCORES_FILE)
            with open(new_scores_path, "w") as wf:
                json.dump({new_file: scores[best_file]}, wf)
            config = Saving.load_dict(cls.config_name, workplace)
            config_bundle = {
                "config": config,
                "device_info": DeviceInfo(cuda, None),
            }
            if config_bundle_callback is not None:
                config_bundle_callback(config_bundle)
            Saving.save_dict(config_bundle, cls.config_bundle_name, pack_folder)
            shutil.copytree(
                os.path.join(workplace, DataModule.package_folder),
                os.path.join(pack_folder, DataModule.package_folder),
            )
        if model_soup_configs is not None:
            m = DLPipeline.load(pack_folder, cuda=cuda, compress=False)
            _generate_model_soup_checkpoint(
                m,
                checkpoint_folder,
                model_soup_configs,
                new_scores_path,
                new_path,
            )
        if compress:
            with lock_manager(base_folder, [pack_folder]):
                Saving.compress(abs_folder, remove_original=True)
        return pack_folder

    @staticmethod
    def get_base(workplace: str) -> Type["DLPipeline"]:
        with open(os.path.join(workplace, DLPipeline.pipeline_file), "r") as f:
            return DLPipeline.get(f.read())  # type: ignore

    @staticmethod
    def load(
        export_folder: str,
        *,
        cuda: Optional[Union[int, str]] = None,
        to_original_device: bool = False,
        compress: bool = True,
        build_trainer: bool = False,
        report: Optional[bool] = None,
        states_callback: states_callback_type = None,
        pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        post_callback: Optional[Callable[["DLPipeline", Dict[str, Any]], None]] = None,
        strict: bool = True,
    ) -> "DLPipeline":
        if export_folder.endswith(".zip"):
            export_folder = export_folder[:-4]
        base_folder = os.path.dirname(os.path.abspath(export_folder))
        with lock_manager(base_folder, [export_folder]):
            with Saving.compress_loader(export_folder, compress):
                pipeline_cls = DLPipeline.get_base(export_folder)
                m = IModifier.load_infrastructure(
                    pipeline_cls,
                    export_folder,
                    None if cuda is None else str(cuda),
                    to_original_device,
                    pre_callback,
                    post_callback,
                )
                try:
                    data_base = DataModule._get_base(export_folder).base
                    m.data_type = data_base.__identifier__
                    loaded = DataModule.load_info(export_folder, return_bytes=True)
                    data_info = loaded.info
                    m.data_module_bytes = loaded.data_module_bytes
                except Exception as err:
                    print_warning(
                        "error occurred when trying to load "
                        f"`DataModule` ({err}), it might cause by BC breaking, "
                        "empty `data_info` will be used"
                    )
                    data_info = {}
                modifier = m._make_modifier()
                modifier.build(data_info, build_trainer=build_trainer, report=report)
                m.model.to(m.device)
                # restore checkpoint
                states = modifier.load_states_from(export_folder)
                modifier.permute_states(states)
                if states_callback is not None:
                    states = states_callback(m, states)
                m.model.load_state_dict(states, strict=strict)
        return m

    def to_onnx(
        self,
        export_folder: str,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        onnx_file: str = "model.onnx",
        opset: int = 11,
        simplify: bool = True,
        forward_fn: Optional[Callable[[Dict[str, Any]], Dict[str, Any]]] = None,
        output_names: Optional[List[str]] = None,
        input_sample: Optional[tensor_dict_type] = None,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> "DLPipeline":
        if input_sample is None:
            if getattr(self, "trainer", None) is None:
                msg = "either `input_sample` or `trainer` should be provided"
                raise ValueError(msg)
            input_sample = self.trainer.input_sample
            input_sample.pop(BATCH_INDICES_KEY, None)
        assert isinstance(input_sample, dict)
        self.model.to_onnx(
            export_folder,
            input_sample,
            dynamic_axes,
            onnx_file=onnx_file,
            opset=opset,
            simplify=simplify,
            forward_fn=forward_fn,
            output_names=output_names,
            num_samples=num_samples,
            verbose=verbose,
            **kwargs,
        )
        return self

    @classmethod
    def pack_onnx(
        cls,
        workplace: str,
        export_folder: str,
        dynamic_axes: Optional[Union[List[int], Dict[int, str]]] = None,
        *,
        input_sample: tensor_dict_type,
        step: Optional[str] = None,
        config_bundle_callback: Optional[Callable[[Dict[str, Any]], Any]] = None,
        pack_folder: Optional[str] = None,
        states_callback: states_callback_type = None,
        pre_callback: Optional[Callable[[Dict[str, Any]], None]] = None,
        post_callback: Optional[Callable[["DLPipeline", Dict[str, Any]], None]] = None,
        onnx_file: str = "model.onnx",
        opset: int = 11,
        simplify: bool = True,
        num_samples: Optional[int] = None,
        verbose: bool = True,
        **kwargs: Any,
    ) -> "DLPipeline":
        packed = cls.pack(
            workplace,
            step=step,
            config_bundle_callback=config_bundle_callback,
            pack_folder=pack_folder,
        )
        m = cls.load(
            packed,
            states_callback=states_callback,
            pre_callback=pre_callback,
            post_callback=post_callback,
        )
        m.to_onnx(
            export_folder,
            dynamic_axes,
            onnx_file=onnx_file,
            opset=opset,
            simplify=simplify,
            input_sample=input_sample,
            num_samples=num_samples,
            verbose=verbose,
            **kwargs,
        )
        return m

    def to_json(self, *, export_path: Optional[str] = None) -> Dict[str, Any]:
        d = shallow_copy_dict(self.config)
        d[self.pipeline_key] = self.__identifier__
        if export_path is not None:
            os.makedirs(os.path.dirname(export_path), exist_ok=True)
            with open(export_path, "w") as f:
                json.dump(d, f)
        return d

    @classmethod
    def from_json(cls, d: Union[str, Dict[str, Any]]) -> "DLPipeline":
        if isinstance(d, dict):
            d = shallow_copy_dict(d)
        elif isinstance(d, str):
            with open(d, "r") as f:
                d = json.load(f)
        else:
            raise ValueError(f"unrecognized input {d} occurred")
        assert isinstance(d, dict)
        name = d.pop(cls.pipeline_key)
        return cls.make(name, d)


def register_modifier(name: str, *, allow_duplicate: bool = False) -> Callable:
    return IModifier.register(name, allow_duplicate=allow_duplicate)


register_modifier("dl")(IModifier)


__all__ = [
    "register_modifier",
    "IModifier",
    "IPipeline",
    "DLPipeline",
]



import os

import numpy as np

from PIL import Image
from PIL import ImageDraw
from PIL import ImageFont
from typing import Any
from typing import Tuple
from cftool.dist import Parallel


tight = False
padding = 0.1
"""
Expected file structure:
-- /path/to/your/fonts/folder
 |- font_type0
  |- font_file0
  |- font_file1
  ...
 |- font_type1
  |- font_file0
  |- font_file1
  ...

A typical use case is to convert opentype / truetype fonts:
-- /path/to/your/fonts/folder
 |- opentype
  |- xxx.otf
  |- xxx.otf
  ...
 |- truetype
  |- xxx.ttf
  |- xxx.ttf
  ...

"""
fonts_folder = "/path/to/your/fonts/folder"

num_jobs = 32
resolution = 512
test_resolution = 32
export_folder = f"export{'-tight' if tight else ''}"

all_folder = os.path.join(export_folder, "all")
split_folder = os.path.join(export_folder, "split")
lower = "abcdefghijklmnopqrstuvwxyz"
upper = "ABCDEFGHIJKLMNOPQRSTUVWXYZ"
alphabet = lower + upper
limit = test_resolution * (1.0 - padding)
width = resolution * 13
height = resolution * 4
os.makedirs(all_folder, exist_ok=True)
os.makedirs(split_folder, exist_ok=True)


def filter_fn(name: str) -> bool:
    if name.startswith("ZillaSlabHighlight"):
        return False
    return True


def get_border(mask: np.ndarray, axis: int) -> int:
    indices = np.argmax(mask, axis=axis)
    first = mask[0] if axis == 0 else mask[:, 0]
    return (indices[indices != 0 | first]).min().item()


def get_bbox(img: np.ndarray) -> Tuple[int, int, int, int]:
    mask = img > 50
    left = get_border(mask, axis=1)
    right = mask.shape[1] - get_border(mask[:, ::-1], axis=1)
    top = get_border(mask, axis=0)
    bottom = mask.shape[0] - get_border(mask[::-1], axis=0)
    return left, right, top, bottom


def get_bbox_from(char: str, font: Any, res: int) -> Tuple[int, int, int, int]:
    image = Image.new("RGB", [res * 2, res * 2])
    draw = ImageDraw.Draw(image)
    draw.text((res, 0), char, font=font)
    left, right, top, bottom = get_bbox(np.array(image.convert("L")))
    return left - res, right - res, top, bottom


def get_font(char: str, font_path: str) -> Tuple[Any, float, float]:
    font_size = 1.0
    while True:
        font = ImageFont.truetype(font_path, int(round(font_size)))
        fw, fh = font.getsize(char)
        if fw >= limit or fh >= limit:
            break
        font_size += 1.0
    left, right, top, bottom = get_bbox_from(char, font, test_resolution)
    ratio = resolution / test_resolution
    if tight:
        span = max(right - left, bottom - top)
        span_ratio = limit / span
        font_size *= span_ratio
        font = ImageFont.truetype(font_path, int(round(font_size)))
        res = int(round(test_resolution * span_ratio))
        left, right, top, bottom = get_bbox_from(char, font, res)
    x_offset = 0.5 * (test_resolution - (left + right))
    y_offset = 0.5 * (test_resolution - (top + bottom))
    font = ImageFont.truetype(font_path, int(round(font_size * ratio)))
    x_offset *= ratio
    y_offset *= ratio
    return font, x_offset, y_offset


def main_task(folder: str, file: str) -> None:
    name = os.path.splitext(file)[0]
    if not filter_fn(name):
        return None
    export_file = f"{name}.png"
    export_path = os.path.join(all_folder, export_file)
    if os.path.isfile(export_path):
        return None
    try:
        x_offsets = set()
        y_offsets = set()
        path = os.path.join(folder, file)
        image = Image.new("RGB", [width, height])
        draw = ImageDraw.Draw(image)
        for i in range(52):
            ix, iy = i % 13, i // 13
            x, y = ix * resolution, iy * resolution
            char = (lower if i < 26 else upper)[i % 26]
            font, x_offset, y_offset = get_font(char, path)
            x_offsets.add(x_offset)
            y_offsets.add(y_offset)
            draw.text((x + x_offset, y + y_offset), char, font=font)
        if max(len(x_offsets), len(y_offsets)) <= 5:
            return None
        image.save(export_path)
    except Exception as err:
        print(f"> failed to export {file} : {err}")


def main() -> None:
    folders = []
    files = []
    for font_type in ["opentype", "truetype"]:
        type_folder = os.path.join(fonts_folder, font_type)
        for file in sorted(os.listdir(type_folder)):
            folders.append(type_folder)
            files.append(file)
    shuffle_indices = np.random.permutation(len(folders))
    folders = [folders[i] for i in shuffle_indices]
    files = [files[i] for i in shuffle_indices]
    Parallel(num_jobs).grouped(main_task, folders, files)


def split_task(file: str) -> None:
    name = os.path.splitext(file)[0]
    name_folder = os.path.join(split_folder, name)
    os.makedirs(name_folder, exist_ok=True)
    img = np.array(Image.open(os.path.join(all_folder, file)))
    for i in range(4):
        for j in range(13):
            char = alphabet[i * 13 + j]
            path = os.path.join(name_folder, f"{char}.png")
            x, y = j * resolution, i * resolution
            ij_img = img[y : y + resolution, x : x + resolution]
            Image.fromarray(ij_img).save(path)


def split() -> None:
    Parallel(num_jobs).grouped(split_task, os.listdir(all_folder))


if __name__ == "__main__":
    main()
    split()



import torch.nn as nn

from torch import Tensor

from .protocols import FFN
from ..convs import DepthWiseConv2d
from ..common import Lambda
from ..customs import Linear
from ..activations import GEGLU
from ..activations import Activation


@FFN.register("ff")
class FeedForward(FFN):
    def __init__(
        self,
        in_dim: int,
        latent_dim: int,
        dropout: float,
        activation: str = "GELU",
        add_last_dropout: bool = True,
    ):
        super().__init__(in_dim, latent_dim, dropout)
        if activation == "geglu":
            blocks = [GEGLU(in_dim, latent_dim)]
        else:
            blocks = [
                Linear(in_dim, latent_dim),
                Activation.make(activation),
            ]
        blocks += [nn.Dropout(dropout), Linear(latent_dim, in_dim)]
        if add_last_dropout:
            blocks.append(nn.Dropout(dropout))
        self.net = nn.Sequential(*blocks)

    @property
    def need_2d(self) -> bool:
        return False

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


@FFN.register("mix_ff")
class MixFeedForward(FFN):
    def __init__(self, in_dim: int, latent_dim: int, dropout: float):
        super().__init__(in_dim, latent_dim, dropout)
        self.net = nn.Sequential(
            Linear(in_dim, latent_dim),
            Lambda(lambda t: t.permute(0, 3, 1, 2), "permute -> BCHW"),
            DepthWiseConv2d(latent_dim),
            Lambda(lambda t: t.flatten(2).transpose(1, 2), "transpose -> BNC"),
            nn.GELU(),
            nn.Dropout(dropout),
            Linear(latent_dim, in_dim),
            nn.Dropout(dropout),
        )

    @property
    def need_2d(self) -> bool:
        return True

    def forward(self, net: Tensor) -> Tensor:
        return self.net(net)


__all__ = [
    "FeedForward",
    "MixFeedForward",
]



import torch.nn as nn

from torch import Tensor
from typing import Any
from typing import Dict
from typing import List
from typing import Optional

from .protocol import VanillaGANMixin
from ...implicit.siren import ImgSiren
from ....misc.internal_.register import register_custom_module
from ....misc.internal_.register import CustomModule


@register_custom_module("siren_gan")
class SirenGAN(VanillaGANMixin, CustomModule):  # type: ignore
    def __init__(
        self,
        img_size: int,
        in_channels: int,
        out_channels: Optional[int] = None,
        latent_dim: int = 128,
        num_classes: Optional[int] = None,
        conditional_dim: int = 16,
        *,
        num_layers: int = 4,
        w_sin: float = 1.0,
        w_sin_initial: float = 30.0,
        bias: bool = True,
        final_activation: Optional[str] = None,
        discriminator: str = "basic",
        discriminator_config: Optional[Dict[str, Any]] = None,
        gan_mode: str = "vanilla",
        gan_loss_config: Optional[Dict[str, Any]] = None,
    ):
        super().__init__()
        self._initialize(
            in_channels=in_channels,
            discriminator=discriminator,
            discriminator_config=discriminator_config,
            num_classes=num_classes,
            gan_mode=gan_mode,
            gan_loss_config=gan_loss_config,
        )
        self.latent_dim = latent_dim
        self.out_channels = out_channels or in_channels
        # siren
        self.siren = ImgSiren(
            img_size,
            self.out_channels,
            latent_dim,
            num_classes,
            conditional_dim,
            num_layers=num_layers,
            w_sin=w_sin,
            w_sin_initial=w_sin_initial,
            bias=bias,
            final_activation=final_activation,
        )

    @property
    def g_parameters(self) -> List[nn.Parameter]:
        return list(self.siren.parameters())

    def decode(
        self,
        z: Tensor,
        *,
        labels: Optional[Tensor],
        size: Optional[int] = None,
        **kwargs: Any,
    ) -> Tensor:
        return self.siren.decode(z, labels=labels, size=size)


__all__ = ["SirenGAN"]
